#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/08 12:15
# @Author  : Wu_RH
# @FileName: OR.py
"""
[AND]与:你可以在后面输入多个左线来表示与关系(题板将按照A规则和B规则)(规则间使用":"隔开)
"""
from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.config.config import PUZZLE_CONFIG

from ...impl_obj import get_rule
from ...summon.solver import Switch

CONFIG = {}
CONFIG.update(PUZZLE_CONFIG)


class RuleOR(AbstractMinesRule):
    name = ["AND", "与"]
    doc = "你可以在后面输入多个左线来表示或关系(题板将按照A规则或B规则)"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        rule_list = [""]
        deep = 0
        for s in data:
            if s == "(":
                if deep > 0:
                    rule_list[-1] += "("
                deep += 1
            elif s == ")":
                if deep > 1:
                    rule_list[-1] += ")"
                deep -= 1
            elif s == ":" and deep == 0:
                rule_list.append("")
            else:
                rule_list[-1] += s
        if len(rule_list) == 0:
            raise ValueError("你不能或空的规则")
        self.rules = []
        for rule in rule_list:
            if CONFIG["delimiter"] in rule:
                rule_name, rule_data = rule.split(CONFIG["delimiter"], 1)
            else:
                rule_name = rule
                rule_data = None
            rule = get_rule(rule_name)(board=board, data=rule_data)
            if not isinstance(rule, AbstractMinesRule):
                continue
            self.rules.append(rule)

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        var_list = []
        for rule in self.rules:
            z = model.NewBoolVar("AND")
            _switch = Switch()
            rule.create_constraints(board=board, switch=_switch)
            model.AddBoolAnd(_switch.get_all_vars()).OnlyEnforceIf(z)
            var_list.append(z)
        model.AddBoolAnd(var_list).OnlyEnforceIf(switch.get(model, self))

    def suggest_total(self, info: dict):
        for rule in self.rules:
            rule.suggest_total(info)



"""
[1M1N] 多雷 + 负雷 + 十字
"""
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.tool import get_logger

def cross_neighbors(pos : AbstractPosition) -> list[AbstractPosition]:
    return [
        pos.up(2),
        pos.down(2),
        pos.left(2),
        pos.right(2),
        pos.up(1),
        pos.down(1),
        pos.left(1),
        pos.right(1)
    ]

class Rule1M1N1X(AbstractClueRule):
    name = ["1M1N1X", "MNX", "多雷 + 负雷 + 十字", "Multiple + Negative + Cross"]
    doc = ""

    def clue_class(self):
        return Value1M1N1X

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        logger = get_logger()
        for pos, _ in board("N"):
            dyed = undyed = 0
            nei = cross_neighbors(pos)
            for t, d in zip(board.batch(nei, mode="type"), board.batch(nei, mode="dye")):
                if (t != "F"):
                    continue
                if d:
                    dyed += 2
                else:
                    undyed += 1
            obj = Value1M1N1X(pos, bytes([abs(dyed - undyed)]))
            board.set_value(pos, obj)
            logger.debug(f"[1M1N]: put {obj} to {pos}")
        return board
    
class Value1M1N1X(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.nei = cross_neighbors(pos)
        self.pos = pos

    def __repr__(self) -> str:
        return str(self.value)

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.nei

    @classmethod
    def type(cls) -> bytes:
        return Rule1M1N1X.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])
    
    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        nei_a = [_pos for _pos in self.nei if board.get_dyed(_pos)] * 2
        nei_b = [_pos for _pos in self.nei if not board.get_dyed(_pos)]

        vars_a = board.batch(nei_a, mode="variable", drop_none=True)
        vars_b = board.batch(nei_b, mode="variable", drop_none=True)

        diff = sum(vars_a) - sum(vars_b)

        # 估计最大绝对值可能为 len(vars_a) + len(vars_b)
        max_abs = len(vars_a) + len(vars_b)
        abs_diff = model.NewIntVar(0, max_abs, "abs_diff")

        model.AddAbsEquality(abs_diff, diff)
        model.Add(abs_diff == self.value).OnlyEnforceIf(s)


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 18:09
# @Author  : Wu_RH
# @FileName: 2Z.py
"""
[2Z] 零和 (Zero-Sum)：每行的染色格与非染色格的雷数相等
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule2Z(AbstractMinesRule):
    name = "2Z", "零和", "Zero-Sum"
    doc = "每行的染色格与非染色格的雷数相等"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        for key in board.get_interactive_keys():
            bound = board.boundary(key=key)
            for _pos in board.get_col_pos(bound):
                line = board.get_row_pos(_pos)
                line_dye = board.batch(line, mode="dye")
                line_var = board.batch(line, mode="variable")
                vars_a = []
                vars_b = []
                for var, dye in zip(line_var, line_dye):
                    if dye:
                        vars_a.append(var)
                    else:
                        vars_b.append(var)
                model.Add(sum(vars_a) == sum(vars_b)).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/03 04:58
# @Author  : Wu_RH
# @FileName: 4F.py
"""
[3F]测试内容: 雷线索表示附近八个格子内的非雷格数
"""
from typing import List, Dict

from ....abs.Mrule import AbstractMinesClueRule, AbstractMinesValue
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.tool import get_logger


class Rule3F(AbstractMinesClueRule):
    name = ["3F", "不是V"]
    doc = "雷线索表示附近八个格子内的非雷格数"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("F"):
            nei_type = board.batch(pos.neighbors(2), mode="type", drop_none=True)
            value = len(nei_type) - nei_type.count("F")
            board.set_value(pos, MinesValue3F(pos, bytes([value])))
        return board


class MinesValue3F(AbstractMinesValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = None):
        self.nei = pos.neighbors(2)
        self.value = code[0]
        self.pos = pos

    def __repr__(self):
        return str(self.value)

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.nei

    @classmethod
    def type(cls) -> bytes:
        return Rule3F.name[0].encode("ascii")

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        logger = get_logger()
        var_list = board.batch(self.nei, mode="variable", drop_none=True)
        model.Add(sum(var_list) == (len(var_list) - self.value)).OnlyEnforceIf(s)
        logger.trace(f"[4F]{self.value}")

    def code(self) -> bytes:
        return bytes([self.value])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/11 21:41
# @Author  : Wu_RH
# @FileName: 2Q.py
"""
[2Q]皇后Queen: 线索表示八方向有雷的方向数。
"""
from typing import List

from minesweepervariants.abs.Rrule import AbstractClueRule, AbstractClueValue
from minesweepervariants.abs.board import AbstractBoard, AbstractPosition
from minesweepervariants.impl.summon.solver import Switch


def pos_shift(
        board: AbstractBoard,
        pos: AbstractPosition,
        index: int
):
    shift = [
        (1, 1), (1, 0), (1, -1),
        (0, 1), (0, -1),
        (-1, 1), (-1, 0), (-1, -1)
    ][index]
    positions = []
    while board.is_valid(_pos := pos.shift(shift[0], shift[1])):
        shift = (
            0 if shift[0] == 0 else (shift[0] + 1 if shift[0] > 0 else shift[0] - 1),
            0 if shift[1] == 0 else (shift[1] + 1 if shift[1] > 0 else shift[1] - 1),
        )
        positions.append(_pos)
    return positions


class Rule2Q(AbstractClueRule):
    name = ["2Q", "皇后", "Queen"]
    doc = "线索表示八方向有雷的方向数量"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            value = 0
            for i in range(8):
                if "F" in board.batch(pos_shift(board, pos, i), mode="type"):
                    value += 1
            board[pos] = Value2Q(pos, bytes([value]))
        return board


class Value2Q(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos, code)
        self.value = code[0]

    def __repr__(self):
        return str(self.value)

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition'] | None:
        positions = []

        for i in range(8):
            pos_list = pos_shift(board, self.pos, i)
            type_list = board.batch(pos_list, mode="type")
            if "F" in type_list:
                positions.append(pos_list[type_list.index("F")])
            else:
                positions.extend(pos_list)

        return positions

    @classmethod
    def type(cls) -> bytes:
        return Rule2Q.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)

        tmp_vars = []
        for index in range(8):
            tmp = model.NewBoolVar("tmp")
            var_list = board.batch(pos_shift(board, self.pos, index), mode="variable")
            model.AddBoolOr(var_list).OnlyEnforceIf(tmp)
            model.Add(sum(var_list) == 0).OnlyEnforceIf(tmp.Not())
            tmp_vars.append(tmp)
        model.Add(sum(tmp_vars) == self.value).OnlyEnforceIf(s)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
# Version:1.1.8
import hashlib
# time:2025.6.3
# from 10:/D:/tool.py

import hashlib
import json
import os
import time
from pathlib import Path
from typing import TextIO
from random import Random

SELF_PATH = os.getcwd()
LOGGER = None
RANDOM = None
SEED = -1


def get_logger(name="logger", log_lv="INFO") -> 'Logger':
    global LOGGER
    if LOGGER is None:
        LOGGER = Logger(name)
        if log_lv == "TRACE":
            LOGGER.print_level = LOGGER.TRACE
        elif log_lv == "DEBUG":
            LOGGER.print_level = LOGGER.DEBUG
        if log_lv == "INFO":
            LOGGER.print_level = LOGGER.INFO
        if log_lv == "WARN":
            LOGGER.print_level = LOGGER.WARN
        if log_lv == "NOTICE":
            LOGGER.print_level = LOGGER.NOTICE
        if log_lv == "WARNING":
            LOGGER.print_level = LOGGER.WARN
        if log_lv == "ERROR":
            LOGGER.print_level = LOGGER.ERROR
        if log_lv == "CRITICAL":
            LOGGER.print_level = LOGGER.CRITICAL
    return LOGGER


def hash_str(s):
    try:
        return int(s)
    except ValueError:
        h = hashlib.sha256(s.encode('utf-8')).hexdigest()
        return int(h[:4], 16)


def get_random(seed: int = -1, new: bool = False) -> Random:
    global RANDOM, SEED
    if RANDOM is None or new:
        if seed == -1:
            seed = int((time.time() * 1e10) % (1e7 + 7))
            seed = hash_str(seed)
        SEED = seed
        get_logger().info("random seed: {}".format(seed))
        RANDOM = Random(seed)
    return RANDOM


class Logger:
    TRACE = 1
    DEBUG = 5
    INFO = 10
    NOTICE = 30
    WARN = 40
    ERROR = 60
    CRITICAL = 100
    time_format = "%Y-%m-%d %H:%M:%S"
    log_root = "log"

    def __init__(self, name: str,
                 lv=10,
                 max_size=262144,
                 log_path=None,
                 log_flag=True
                 ):
        self.print_level = lv
        self.max_size = max_size
        self.name = name
        self.log_flag = log_flag
        self.log_path = os.path.join(SELF_PATH, self.log_root) \
            if log_path is None else log_path

        self.file_id = 0
        self.file = None

        self.start()

    def __del__(self):
        self.close()

    def get_time(self):
        return time.strftime(self.time_format, time.localtime())

    def __create_file(self):
        if not os.path.exists(self.log_root):
            os.makedirs(self.log_root)
        file_name = os.path.join(self.log_root,
                                 f"{self.name}_{self.file_id}.log")
        while 1:
            if not Path(file_name).exists():
                open(file_name, 'x').close()
                break
            else:
                if os.path.getsize(file_name) > self.max_size:
                    self.file_id += 1
                    file_name = os.path.join(self.log_root, "{}_{}.log"
                                             .format(self.name, self.file_id))
                    continue
                else:
                    break
        self.file = open(file_name, 'a', encoding='utf-8')

    def __log(self, log_type, msg, log_lv, end="\n"):
        if self.print_level > log_lv:
            return
        s = f"<{self.get_time()}>" + f"[{log_type}]:" + f'{msg}{end}'
        print(s, end="", flush=True)
        self.file.write(s)
        self.file.flush()

        if (self.max_size != -1 and
                os.path.getsize(self.file.name) > self.max_size):
            self.__create_file()

    def start(self):
        if self.file is None or self.file.closed:
            self.__create_file()
            if self.log_flag:
                self.__log("INFO", f"{self.name} log start", 4)

    def close(self):
        if not self.file.closed:
            if self.log_flag:
                self.__log("INFO", f"{self.name} log end\n", 4)
            self.file.close()

    def trace(self, msg, end="\n"):
        self.__log("TRACE", msg, log_lv=self.TRACE)

    def debug(self, msg, end="\n"):
        self.__log("DEBUG", msg, end=end, log_lv=self.DEBUG)

    def info(self, msg, end="\n"):
        self.__log("INFO", msg, end=end, log_lv=self.INFO)

    def warning(self, msg, end="\n"):
        self.__log("WARN", msg, end=end, log_lv=self.WARN)

    def warn(self, msg, end="\n"):
        self.__log("WARN", msg, end=end, log_lv=self.WARN)

    def notice(self, msg, end="\n"):
        self.__log("NOTICE", msg, end=end, log_lv=self.NOTICE)

    def error(self, msg, end="\n"):
        self.__log("ERROR", msg, end=end, log_lv=self.ERROR)

    def critical(self, msg, end="\n"):
        self.__log("CRITICAL", msg, end=end, log_lv=self.CRITICAL)


class GetData:
    def __init__(self, data_name, encoding="utf-8", data_path="data"):
        self.encoding = encoding
        self.data_name = data_name
        self.data_path = data_path
        self.io = None
        self.data = json.load(self.get_io())

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __load_io(self, mod="r"):
        return open(os.path.join(SELF_PATH, self.data_path, self.data_name + ".json"),
                    mod, encoding=self.encoding)

    def get_io(self, mod="r") -> TextIO:
        if self.io is None:
            self.io = self.__load_io(mod)
        if self.io is not None and self.io.mode != mod:
            self.io.close()
            self.io = self.__load_io(mod)
        return self.io

    def reload_data(self):
        self.io.close()
        self.io = None
        self.data = json.load(self.get_io("r"))

    def update_data(self):
        json.dump(self.get_io("w"), self.data, ensure_ascii=False, indent=4)
        self.io.close()
        self.io = None



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/21 08:12
# @Author  : Wu_RH
# @FileName: 3D1B.py
"""
[3D1B']平衡: 所有平行面的总雷数均相等
"""
from math import gcd
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.impl.summon.solver import Switch
from .. import Abstract3DMinesRule


class Rule(Abstract3DMinesRule):
    name = ["3D1B'", "平衡"]
    doc = "所有平行面的总雷数均相等"

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s1 = switch.get(model, self, "↔")
        s2 = switch.get(model, self, "↕")
        s3 = switch.get(model, self, "X")

        # 层间面平衡
        _var_list = None
        for key in board.get_interactive_keys():
            var_list = [var for pos, var in board(mode="var", key=key)]
            if _var_list is None:
                _var_list = var_list
                continue
            model.Add(sum(var_list) == sum(_var_list)).OnlyEnforceIf(s3)
            _var_list = var_list

        # col轴面平衡
        _var_list = None
        for pos in board.get_row_pos(board.boundary()):
            var_list = []
            for key in board.get_interactive_keys():
                _pos = pos.clone()
                _pos.board_key = key
                var_list.extend(board.batch(
                    board.get_col_pos(_pos),
                    mode="var"
                ))
            if _var_list is None:
                _var_list = var_list
                continue
            model.Add(sum(var_list) == sum(_var_list)).OnlyEnforceIf(s2)
            _var_list = var_list

        # row轴面平衡
        _var_list = None
        for pos in board.get_col_pos(board.boundary()):
            var_list = []
            for key in board.get_interactive_keys():
                _pos = pos.clone()
                _pos.board_key = key
                var_list.extend(board.batch(
                    board.get_row_pos(_pos),
                    mode="var"
                ))
            if _var_list is None:
                _var_list = var_list
                continue
            model.Add(sum(var_list) == sum(_var_list)).OnlyEnforceIf(s1)
            _var_list = var_list

    def suggest_total(self, info: dict):
        def lcm(a, b):
            # 计算两个数的最小公倍数
            if a == 0 or b == 0:
                return 0
            return a * b // gcd(a, b)

        def lcm3(a, b, c):
            # 计算三个数的最小公倍数
            return lcm(a, lcm(b, c))

        def add_constraints(model, total_var):
            nonlocal base
            # 计算乘数的上限：总雷数的最大值除以base
            max_multiplier = total_var.Proto().domain[1] // base + 1
            mult = model.NewIntVar(0, max_multiplier, f"mult")
            model.AddMultiplicationEquality(total_var, [mult, base])

        # 获取每个交互区域的尺寸，假设每个尺寸是三元组 (w, h, d)
        sizes = [info["size"][interactive] for interactive in info["interactive"]]
        d = len(info["interactive"])
        w, h = sizes[0][0], sizes[0][1]
        for size in sizes[1:]:
            if w != size[0] or h != size[1]:
                raise ValueError("保证题板尺寸相同")
        if w > 0 and h > 0 and d > 0:  # 确保所有尺寸为正
            base = lcm3(w, h, d)

        # 如果没有有效的base，则跳过添加约束
        info["hard_fns"].append(add_constraints)


"""
[2K] 金将 (Kinsho)：线索表示金将范围（即上、下、左、右、左上、右上六格）内的雷数
"""

from minesweepervariants.utils.impl_obj import MINES_TAG, VALUE_QUESS
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.tool import get_logger

class Rule2K(AbstractClueRule):
    name = ["2K", "金将", "Kinsho"]
    doc = "线索表示金将范围（即上、下、左、右、左上、右上六格）内的雷数"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        logger = get_logger()
        for pos, _ in board("N"):
            neighbors = self.get_kinsho_neighbors(pos, board)
            mine_count = sum(1 for n in neighbors if board.get_type(n) == "F")
            board.set_value(pos, Value2K(pos, count=mine_count))
        return board

    @staticmethod
    def get_kinsho_neighbors(pos: AbstractPosition, board: AbstractBoard):
        x, y = pos.x, pos.y
        board_key = pos.board_key
        directions = [(0, 1), (0, -1), (1, 0), (-1, 0), (-1, 1), (-1, -1)]
        neighbors = []
        for dx, dy in directions:
            npos = type(pos)(x + dx, y + dy, board_key)
            if board.in_bounds(npos):
                neighbors.append(npos)
        return neighbors

class Value2K(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            self.count = code[0]
        else:
            self.count = count

    def __repr__(self):
        return str(self.count)

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return Rule2K.get_kinsho_neighbors(self.pos, board)

    @classmethod
    def type(cls) -> bytes:
        return b'2K'

    def code(self) -> bytes:
        return bytes([self.count])
    
    def deduce_cells(self, board: 'AbstractBoard') -> bool:
        type_dict = {"N": [], "F": []}
        for pos in self.high_light(board):
            t = board.get_type(pos)
            if t in ("", "C"):
                continue
            type_dict[t].append(pos)
        n_num = len(type_dict["N"])
        f_num = len(type_dict["F"])
        if n_num == 0:
            return False
        if f_num == self.count:
            for i in type_dict["N"]:
                board.set_value(i, VALUE_QUESS)
            return True
        if f_num + n_num == self.count:
            for i in type_dict["N"]:
                board.set_value(i, MINES_TAG)
            return True
        return False
    
    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self.pos)
        neighbor_vars = []
        for neighbor in self.high_light(board):
            var = board.get_variable(neighbor)
            if var is not None:
                neighbor_vars.append(var)
        if neighbor_vars:
            model.Add(sum(neighbor_vars) == self.count).OnlyEnforceIf(s)
            get_logger().trace(f"[2K] {self.pos}: {self.count} constraint: {[str(n) for n in neighbor_vars]} == {self.count}")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/05 07:51
# @Author  : Wu_RH
# @FileName: 2P.py
"""
[2P] 乘积 (Product)：线索表示距离最近的 2 个雷的距离之积
"""
from typing import Dict

from ....abs.Rrule import AbstractClueValue, AbstractClueRule
from ....abs.board import AbstractPosition, AbstractBoard
from ....utils.image_create import get_text, get_image, get_row, get_col, get_dummy
from ....utils.tool import get_logger


def sqrt_form(n):
    """
    将正整数 n 写作n平方根的最简形式：a√b。
    例如：输入 8，输出 "2√2"；输入 4，输出 "2"；输入 2，输出 "√2"。
    """
    if n <= 0:
        return 0, -1

    a = 1  # 初始化平方因子部分
    b = n  # 初始化剩余部分
    i = 2  # 从最小质因数开始遍历

    # 提取所有平方因子：将 n 分解为 a^2 * b（b 无平方因子）
    while i * i <= b:
        # 当 i^2 整除 b 时，重复除去该因子
        while b % (i * i) == 0:
            b //= i * i  # 整数除法，更新剩余部分
            a *= i  # 更新平方因子部分
        i += 1

    # 根据 a 和 b 的值格式化输出
    if b == 1:
        return a, -1  # 例如 n=4, 输出 "2"
    elif a == 1:
        return -1, b  # 例如 n=2, 输出 "√2"
    else:
        return a, b  # 例如 n=8, 输出 "2√2"


def get_factor_pairs(n):
    """
    返回正整数n的所有正整数乘法因子对（去重）

    参数:
        n: 正整数

    返回:
        列表，包含所有因子对 (a, b) 满足 a × b = n 且 a ≤ b
        按第一个因子升序排列
    """
    pairs = []
    # 只需遍历到平方根即可
    for i in range(1, int(n ** 0.5) + 1):
        if n % i == 0:
            j = n // i
            # 确保 a ≤ b
            if i <= j:
                pairs.append((i, j))
    return pairs


class Rule2P(AbstractClueRule):
    name = ["2P", "乘积", "Product"]
    doc = "线索表示距离最近的 2 个雷的距离之积"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        if len([_ for _ in board("F")]) < 2:
            return board
        for pos, _ in board("N"):
            r = 0
            a_lay = b_lay = -1

            while b_lay == -1:
                r += 1
                count = board.batch(pos.neighbors(r, r), mode="type").count("F")
                if count >= 2:
                    if a_lay == -1:
                        a_lay = b_lay = r
                    else:
                        b_lay = r
                elif count == 1:
                    if a_lay == -1:
                        a_lay = r
                    else:
                        b_lay = r
            if a_lay * b_lay > 254:
                value = a_lay * b_lay
                obj = Value2P(pos, bytes([value // 255, value % 255]))
            else:
                obj = Value2P(pos, bytes([a_lay * b_lay]))
            board.set_value(pos, obj)
        return board


class Value2P(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.pos = pos
        if len(code) == 1:
            self.value = code[0]
        elif len(code) == 2:
            self.value = code[0] * 255 + code[1]

    def __repr__(self):
        value_a, value_b = sqrt_form(self.value)
        if value_b == -1:
            return f"{value_a}"
        elif value_a == -1:
            return f"√{value_b}"
        return f"{value_a}√{value_b}"

    def web_component(self, board) -> Dict:
        value_a, value_b = sqrt_form(self.value)
        if value_b == -1:
            return get_text(str(value_a))
        if value_a == -1:
            return get_text(
                "$\\sqrt{" + str(value_b) + "}$"
            )
        else:
            return get_text(
                "$" + str(value_a) +
                "\\sqrt{" + str(value_b) +
                "}$"
            )

    def compose(self, board) -> Dict:
        value_a, value_b = sqrt_form(self.value)
        if value_b == -1:
            return get_col(
                get_dummy(height=0.175),
                get_text(str(value_a)),
                get_dummy(height=0.175),
            )
        elif value_a == -1:
            return get_row(
                get_image("sqrt"),
                get_text(str(value_b)),
                spacing=-0.15
            )
        else:
            return get_row(
                get_text(str(value_a)),
                get_image("sqrt"),
                get_text(str(value_b)),
                spacing=-0.2
            )

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        n = 1
        v = 0
        positions = []
        while True:
            neibors = self.pos.neighbors(n, n)
            neibors = [pos for pos in neibors if board.in_bounds(pos)]
            if len(neibors) == 0:
                return []
            neibors_t = board.batch(neibors, mode="type")
            v += neibors_t.count("F") + neibors_t.count("N")
            for pos, t in zip(neibors, neibors_t):
                if t not in ["F", "N"]:
                    continue
                positions.append(pos)
            if v >= 2:
                break
            n += 1
        return positions

    @classmethod
    def type(cls) -> bytes:
        return Rule2P.name[0].encode("ascii")

    def code(self) -> bytes:
        if self.value > 254:
            return bytes([self.value // 255, self.value % 255])
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        logger = get_logger()
        s = switch.get(model, self)

        var_list = []
        for par_a, par_b in get_factor_pairs(self.value):
            var = model.NewBoolVar("[2P]")
            if par_a == par_b:
                nei = self.pos.neighbors(par_a, par_a)
                model.Add(sum(board.batch(nei, mode="variable", drop_none=True)) >= 2).OnlyEnforceIf([var, s])
                none_var = self.pos.neighbors(1, par_a - 1)
                model.Add(sum(board.batch(none_var, mode="variable", drop_none=True)) == 0).OnlyEnforceIf([var, s])
            else:
                nei_a = self.pos.neighbors(par_a, par_a)
                model.Add(sum(board.batch(nei_a, mode="variable", drop_none=True)) == 1).OnlyEnforceIf([var, s])
                nei_b = self.pos.neighbors(par_b, par_b)
                model.Add(sum(board.batch(nei_b, mode="variable", drop_none=True)) >= 1).OnlyEnforceIf([var, s])
                none_var = self.pos.neighbors(0, par_a - 1) + self.pos.neighbors(par_a + 1, par_b - 1)
                model.Add(sum(board.batch(none_var, mode="variable", drop_none=True)) == 0).OnlyEnforceIf([var, s])
            var_list.append(var)
        model.AddBoolOr(var_list).OnlyEnforceIf(s)
        logger.trace(f"pos {self.pos} value[{self}] 添加了共{len(var_list)}情况")



from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard, AbstractPosition

class Rule4N(AbstractMinesRule):
    name = ["4N", "4N", "相邻"]
    doc = "非雷周围四格必须有雷"

    def create_constraints(self, board, switch):
        model = board.get_model()
        s = switch.get(model, self)

        for pos, var in board(mode="variable"):
            model.AddBoolOr(board.batch(pos.neighbors(1), mode="variable", drop_none=True)).OnlyEnforceIf([var.Not(), s])



"""
[1L1M] 误差 + 多雷
"""

from typing import List

from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition

from ....utils.tool import get_logger, get_random
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG

class Rule1L1M(AbstractClueRule):
    name = ["1L1M", "LM", "误差 + 多雷", "Liar + Multiple"]
    doc = ""
    
    def fill(self, board: AbstractBoard) -> AbstractBoard:
        random = get_random()
        logger = get_logger()
        for pos, _ in board("N"):
            positions = pos.neighbors(2)
            value = 0
            for t, d in zip(
                    board.batch(positions, "type"),
                    board.batch(positions, "dye")
            ):
                if t != "F":
                    continue
                if d:
                    value += 2
                else:
                    value += 1
            value += 1 if random.random() > 0.5 else -1
            if (value < 0): # 0 - 1
                value = 1
            board.set_value(pos, Value1L1M(pos, code=bytes([value])))
            logger.debug(f"[1L1M]: put {value} to {pos}")
        return board
    
class Value1L1M(AbstractClueValue):
    value: int
    neighbors: list

    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos)
        self.value = code[0]
        self.neighbors = pos.neighbors(2)

    def __repr__(self) -> str:
        return f"{self.value}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbors
    
    @classmethod
    def type(cls) -> bytes:
        return Rule1L1M.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])
    
    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        vals = []
        offset = 0
        dyes = board.batch(self.neighbors, "dye")
        for pos, dye in zip(self.neighbors, dyes):
            if board.get_type(pos) == "C":
                continue
            if board.get_type(pos) == "F":
                offset += 2 if dye else 1
                continue
            if not board.in_bounds(pos):
                continue
            if dye:
                vals.append(board.get_variable(pos) * 2)
            else:
                vals.append(board.get_variable(pos))
        if vals:
            neighbor_sum = sum(vals)
            b1 = model.NewBoolVar("sum_eq_count_plus_1")
            b2 = model.NewBoolVar("sum_eq_count_minus_1")
            remaining = self.value - offset

            model.Add(neighbor_sum == remaining + 1).OnlyEnforceIf(b1)
            model.Add(neighbor_sum != remaining + 1).OnlyEnforceIf(b1.Not())
            model.Add(neighbor_sum == remaining - 1).OnlyEnforceIf(b2)
            model.Add(neighbor_sum != remaining - 1).OnlyEnforceIf(b2.Not())

            model.AddBoolOr([b1, b2]).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/10 18:09
# @Author  : Wu_RH
# @FileName: 1M.py
"""
[1M]多雷: 每个染色格的雷被视为两个(总雷数不受限制)
"""

from ....abs.board import AbstractBoard, AbstractPosition
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....utils.tool import get_logger


class Rule1M(AbstractClueRule):
    name = ["1M", "M", "多雷"]
    doc = "每个染色格的雷被视为两个(总雷数不受限制)"

    def fill(self, board: 'AbstractBoard'):
        logger = get_logger()
        for pos, _ in board("N"):
            positions = pos.neighbors(2)
            value = 0
            for t, d in zip(
                    board.batch(positions, "type"),
                    board.batch(positions, "dye")
            ):
                if t != "F":
                    continue
                if d:
                    value += 2
                else:
                    value += 1
            obj = Value1M(pos, code=bytes([value]))
            board.set_value(pos, obj)
            logger.debug(f"[1M]: put {obj} to {pos}")
        return board


class Value1M(AbstractClueValue):
    value: int
    neighbors: list

    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos)
        self.value = code[0]
        self.neighbors = pos.neighbors(2)

    def __repr__(self) -> str:
        return f"{self.value}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbors

    @classmethod
    def type(cls) -> bytes:
        return Rule1M.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        vals = []
        offset = 0
        dyes = board.batch(self.neighbors, "dye")
        for pos, dye in zip(self.neighbors, dyes):
            if board.get_type(pos) == "C":
                continue
            if board.get_type(pos) == "F":
                offset += 2 if dye else 1
                continue
            if not board.in_bounds(pos):
                continue
            if dye:
                vals.append(board.get_variable(pos) * 2)
            else:
                vals.append(board.get_variable(pos))
        if vals:
            model.Add(sum(vals) == (self.value - offset)).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/29 10:00
# @Author  : AI Assistant
# @FileName: 3DW.py

"""
[3DW]海浪(Wave): 所有层的同一行列中有且只有一格有雷, 且所有雷的水平邻居中每一个的上下偏移一格以内的范围内必定有雷.
1. 对于任意x,y, 存在唯一z使得(x,y,z)有雷
2. 若(x0,y0,z0)有雷, 则{(x, y, z)| |x - x0| <= 1, |y - y0| <= 1, |z - z0| <= 1}中的雷格数为其2D邻格个数+1
"""

from typing import List

from .. import Abstract3DMinesRule
from .....abs.board import AbstractPosition, AbstractBoard


def get_vertical_column(board: AbstractBoard, pos: AbstractPosition) -> List[AbstractPosition]:
    """获取给定x,y坐标在所有层中的位置列表"""
    positions = []
    keys = board.get_interactive_keys()

    for key in keys:
        # 克隆位置并改变层
        new_pos = pos.clone()
        new_pos.board_key = key
        if board.in_bounds(new_pos):
            positions.append(new_pos)

    return positions


def get_3d_neighbors(board: AbstractBoard, center_pos: AbstractPosition) -> List[AbstractPosition]:
    """获取3D空间中的27邻域(3x3x3立方体)内的所有位置"""
    neighbors = []

    # 遍历3x3x3立方体
    for dx in [-1, 0, 1]:
        for dy in [-1, 0, 1]:
            for dz in [-1, 0, 1]:
                # 计算新位置
                temp_pos = center_pos.clone()

                # x轴移动
                if dx > 0:
                    temp_pos = temp_pos.right()
                elif dx < 0:
                    temp_pos = temp_pos.left()

                if temp_pos is None:
                    continue

                # y轴移动
                if dy > 0:
                    temp_pos = temp_pos.down()
                elif dy < 0:
                    temp_pos = temp_pos.up()

                if temp_pos is None:
                    continue

                # z轴移动
                if dz > 0:
                    temp_pos = Rule3DW.up(board, temp_pos, 1)
                elif dz < 0:
                    temp_pos = Rule3DW.down(board, temp_pos, 1)

                if temp_pos is not None and board.in_bounds(temp_pos):
                    neighbors.append(temp_pos)

    return neighbors


class Rule3DW(Abstract3DMinesRule):
    name = ["3DW", "海浪", "Wave"]
    doc = """所有层的同一行列中有且只有一格有雷, 且所有雷的水平邻居中每一个的上下偏移一格以内的范围内必定有雷.
1. 对于任意x,y, 存在唯一z使得(x,y,z)有雷
2. 若(x0,y0,z0)有雷, 则{(x, y, z)| |x - x0| <= 1, |y - y0| <= 1, |z - z0| <= 1}中的雷格数为其2D邻格个数+1"""

    def suggest_total(self, info: dict):
        # 建议合适的雷数
        size = next(iter(info["size"].values()))
        total_cells = size[0] * size[1]
        info["hard_fns"].append(lambda model, total: model.Add(total == int(total_cells)))
        info["soft_fn"](total_cells, 3)

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s1 = switch.get(model, self)
        s2 = switch.get(model, self)

        boundary = board.boundary()
        for col_pos in board.get_col_pos(boundary):
            for pos in board.get_row_pos(col_pos):
                # 获取该(x,y)位置在所有层的垂直柱
                column_positions = get_vertical_column(board, pos)
                if column_positions:
                    var_list = [board.get_variable(p) for p in column_positions]
                    # 每个垂直柱中有且只有一个雷
                    model.Add(sum(var_list) == 1).OnlyEnforceIf(s1)

        boundary = board.boundary()
        for col_pos in board.get_col_pos(boundary):
            for pos in board.get_row_pos(col_pos):
                # 对每一层的每个位置检查
                for key in board.get_interactive_keys():
                    center_pos = pos.clone()
                    center_pos.board_key = key
                    if not board.in_bounds(center_pos):
                        continue

                    # 获取3x3x3邻域
                    neighbors = get_3d_neighbors(board, center_pos)

                    neighbors2d = [neighbor for neighbor in center_pos.neighbors(2) if board.in_bounds(neighbor)]

                    center_var = board.get_variable(center_pos)
                    neighbor_vars = [board.get_variable(neighbor) for neighbor in neighbors]

                    model.Add(sum(neighbor_vars) == len(neighbors2d)+1).OnlyEnforceIf([center_var, s2])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/11 17:06
# @Author  : Wu_RH
# @FileName: 4C.py
"""
[4C]十字路口(Crossing): 雷区域可以分为6组不重合的八连通蛇，使得对于题板任意两个边都可以找到1个雷组满足只与这两条边接触。
"""
from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.impl.summon.solver import Switch


class Rule4C(AbstractMinesRule):
    # name = ["4C", "十字路口", "Crossing"]

    def suggest_total(self, info: dict):
        def a(model, total):
            model.Add(total >= 12)
        info["hard_fns"].append(a)

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        ...






#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/06 02:11
# @Author  : QuirkyStorm7988
# @FileName: 1W.py
"""
[1P] 分组 (Partition)：线索表示 3x3 范围内连续雷的组数
"""
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractPosition, AbstractBoard
from ....utils.tool import get_logger


def MineStatus_1P(clue: int) -> list[int]:
    """
    返回值：一个int列表，其中存的每一个int表示：
        一个二进制数，第i位（从低到高）表示从左上角开始顺时针旋转，第i个格子的雷情况（是雷->1，非雷->0）
        将这个二进制数转化为十进制存储到元素当中，如42(10) == 00101010(2)，即这个线索格的右上、右下、左下有雷
    """
    ans = []
    a = [0 for _ in range(8)]  # 决策列表

    def dfs(step: int):
        if step >= 8:  # 最终处理
            # 先写没有剪枝的
            test = []
            last = 0
            for i in range(8):
                if a[i]:
                    last += 1
                else:
                    if last != 0: test.append(last)
                    last = 0
            if last != 0: test.append(last)
            if a[-1] and a[0] and len(test) != 1:
                test[0] += test[-1]
                del test[-1]
            if len(test) != clue: return None
            #
            status = 0
            for i in range(8):
                status += 2 ** i * a[i]
            if status not in ans:
                ans.append(status)
            # if a[:] not in ans:
            #     ans.append(a[:])
            return None
        a[step] = 0
        dfs(step + 1)
        a[step] = 1
        dfs(step + 1)
        return None

    dfs(0)
    return ans


class Rule1P(AbstractClueRule):
    name = ["1P", "P", "分组", "Partition"]
    doc = "线索表示 3x3 范围内连续雷的组数"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        logger = get_logger()
        for pos, _ in board("N"):
            nei = [pos.right(), pos.right().down(), pos.down(), pos.left().down(),
                   pos.left(), pos.left().up(), pos.up(), pos.right().up()]
            nei_type = board.batch(nei, mode="type")
            value = 0
            _t = ""
            for t in nei_type:
                if t != "F" and _t == "F":
                    value += 1
                _t = t
            if nei_type[-1] == "F" and nei_type[0] != "F":
                value += 1
            obj = Value1P(pos, bytes([value]))
            board.set_value(pos, obj)
            logger.debug(f"[1P]set {obj} to {pos}")

        return board


class Value1P(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.pos = pos

    def __repr__(self):
        return str(self.value)

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.pos.neighbors(2)

    @classmethod
    def type(cls) -> bytes:
        return Rule1P.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        var_list = board.batch([
            self.pos.right(), self.pos.right().down(),
            self.pos.down(), self.pos.left().down(),
            self.pos.left(), self.pos.left().up(),
            self.pos.up(), self.pos.right().up()
        ], mode="variable")

        possible_list = [[]]

        for value in MineStatus_1P(self.value):
            bool_list = [(value >> i) & 1 == 1 for i in reversed(range(8))]
            flag = False
            for index, var in enumerate(var_list):
                if var is None and bool_list[index]:
                    flag = True
                    break
                if var is None:
                    continue
                possible_list[-1].append(bool_list[index])
            if flag:
                possible_list.pop(-1)
            possible_list.append([])

        if any(v is None for v in var_list):
            var_list = [var for var in var_list if var is not None]
        possible_list.pop(-1)

        model.AddAllowedAssignments(var_list, possible_list).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/15 18:07
# @Author  : Wu_RH
# @FileName: generate_puzzle.py.py
import base64
import os
import time

from minesweepervariants.impl.impl_obj import get_board, ModelGenerateError, encode_board
from minesweepervariants.impl.summon import Summon
from minesweepervariants.impl.summon.summon import GenerateError
from minesweepervariants.utils.image_create import draw_board
from minesweepervariants.utils.tool import get_logger, get_random
from minesweepervariants.utils.impl_obj import get_seed

from minesweepervariants.config.config import DEFAULT_CONFIG, PUZZLE_CONFIG

# ==== 获取默认值 ====
CONFIG = {}
CONFIG.update(DEFAULT_CONFIG)
CONFIG.update(PUZZLE_CONFIG)


def main(
        log_lv: str,  # 日志等级
        seed: int,  # 随机种子
        attempts: int,  # 尝试次数
        size: tuple[int, int],  # 题板尺寸
        total: int,  # 总雷数
        rules: list[str],  # 规则id集合
        dye: str,  # 染色规则
        mask_dye: str,   # 异形题板
        drop_r: bool,  # 在推理时候是否隐藏R推理
        board_class: str,  # 题板的名称
        vice_board: bool,  # 启用删除副板
        unseed: bool,
):
    rule_code = rules[:]
    logger = get_logger(log_lv=log_lv)
    get_random(seed, new=True)
    attempts = 20 if attempts == -1 else attempts
    s = Summon(size=size, total=total, rules=rules, board=board_class,
               drop_r=drop_r, mask=mask_dye, dye=dye, vice_board=vice_board)
    if unseed:
        s.unseed = True
    total = s.total
    logger.info(f"total mines: {total}")
    _board = None
    info_list = []
    print(rule_code)
    attempt_index = 0
    while attempts == -1 or attempt_index < attempts:
        attempt_index += 1
        s = Summon(size=size, total=total, rules=rule_code[:], board=board_class,
                   drop_r=drop_r, mask=mask_dye, dye=dye, vice_board=vice_board)
        if unseed:
            s.unseed = True
        logger.info(f"尝试第{attempt_index}次minesweepervariants..", end="\r")
        get_random(seed, new=True)
        a_time = time.time()
        try:
            _board = s.create_puzzle()
        except ModelGenerateError:
            continue
        except GenerateError:
            continue
        b_time = time.time()
        if _board is None:
            continue
        n_num = len([None for _ in _board("N")])
        logger.info(f"<{attempt_index}>生成用时:{b_time - a_time}s")
        logger.info(f"总雷数: {total}/{n_num}")
        logger.info("\n" + _board.show_board())
        if len([None for _ in _board("NF")]) == total:
            logger.warn("题板生成失败 线索填充无法覆盖全盘")
            continue
        info_list.append([
            b_time - a_time,
            n_num,
            "\n" + _board.show_board(),
            _board.encode(),
            "\n" + s.answer_board_str,
            s.answer_board_code,
            _board
        ])

        break

    if not info_list:
        raise ValueError("未在有效次数内得出结果")

    info_list.sort(key=lambda x: x[0])
    time_used, n_num, board_str, board_code, answer, answer_code, _board = info_list[0]

    rule_text = ""
    for rule in rules:
        rule_text += "[" + rule + "]"
    if rule_text == "":
        rule_text = "[V]"
    if dye:
        rule_text += f"[@{dye}]"
    if mask_dye:
        rule_text += f"[{mask_dye}]"
    size_a = 0
    size_b = 0
    size_c = len(_board.get_interactive_keys())
    for key in _board.get_interactive_keys():
        bound = _board.boundary(key)
        size_a = max(size_a, bound.x + 1)
        size_b = max(size_b, bound.y + 1)
    rule_text += f"{size_a}x{size_b}"
    if size_c > 1:
        rule_text += f"x{size_c}"

    if not os.path.exists(CONFIG["output_path"]):
        os.mkdir(CONFIG["output_path"])

    i, mask = 1, 0
    for key in _board.get_board_keys():
        for _, obj in _board(key=key):
            if obj is None:
                mask += i
            i <<= 1

    # 计算需要的字节长度
    byte_length = (mask.bit_length() + 7) // 8  # 计算所需字节数
    byte_length = max(byte_length, 1)  # 确保至少 1 字节

    mask = mask.to_bytes(byte_length, "big", signed=False)

    with open(os.path.join(CONFIG["output_path"], "demo.txt"), "a", encoding="utf-8") as f:
        f.write("\n" + ("=" * 100) + "\n\n生成时间" + logger.get_time() + "\n")
        f.write(f"生成用时:{time_used}s\n")
        f.write(f"总雷数: {total}/{n_num}\n")
        f.write(f"种子: {get_seed()}\n")
        f.write(rule_text)
        f.write(board_str)
        f.write(answer)

        f.write(f"\n答案: img -c {encode_board(answer_code)} ")
        f.write(f"-r \"{rule_text}-R{total}/")
        f.write(f"{n_num}")
        if unseed:
            f.write(f"-{get_seed()}\" ")
        else:
            f.write(" ")
        f.write("-o answer\n")

        f.write(f"\n题板: img -c {encode_board(board_code)} ")
        f.write(f"-r \"{rule_text}-R{'*' if drop_r else total}/")
        f.write(f"{n_num}")
        if unseed:
            f.write(f"-{get_seed()}\" ")
        else:
            f.write(" ")
        f.write("-o demo\n")

        rule_code = [base64.urlsafe_b64encode(rule.encode("utf-8")).decode("utf-8") for rule in rule_code]
        f.write(f"\n题板代码: \n{encode_board(answer_code)}:{mask.hex()}:{':'.join(rule_code)}\n")

    image_bytes = draw_board(board=get_board(board_class)(code=board_code), cell_size=100, output="demo",
                             bottom_text=(rule_text +
                                          f"-R{'*' if drop_r else total}/{n_num}" +
                                          ("\n" if unseed else f"-{get_seed()}\n")))
    draw_board(board=get_board(board_class)(code=answer_code), output="answer", cell_size=100,
               bottom_text=(rule_text +
                            f"-R{total}/{n_num}" +
                            ("\n" if unseed else f"-{get_seed()}\n")))

    filepath = os.path.join(CONFIG["output_path"], "demo.png")
    with open(filepath, "wb") as f:
        f.write(image_bytes)

    logger.info("\n\n" + "=" * 20 + "\n")
    logger.info("\n生成时间" + logger.get_time() + "\n")
    logger.info(f"生成用时:{time_used}s\n")
    logger.info(f"总雷数: {total}/{n_num}\n")
    logger.info(board_str + "\n")
    logger.info(answer + "\n")
    logger.info(f"{board_code}")





from minesweepervariants.server.model import Model


class MPModel(Model):
    def __init__(self, host: Model | None = None, token: str = '',  *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.players: list[Model] = []
        self.token = token
        self.host = host
        print(f"MPModel initialized with token: {self.token}, host: {self.host}")
        if host is not None:
            if hasattr(host, "add_player") and callable(getattr(host, "add_player")):
                getattr(host, "add_player")(self)


    def add_player(self, player: Model):
        self.players.append(player)

    def remove_player(self, player: Model):
        self.players.remove(player)

    def generate_board(self):
        from minesweepervariants.impl.summon.game import GameSession

        if self.host is None:
            return super().generate_board()

        try:
            host_game = self.host.get_game()
        except Exception as e:
            return {"reason": f"主机未准备好: {e}", "success": False}, 200

        try:
            _ = host_game.board
            _ = host_game.answer_board
        except Exception as e:
            return {"reason": f"主机棋盘未准备好: {e}", "success": False}, 200


        summon_arg = getattr(self.host, "summon", None) or getattr(host_game, "summon", None)
        if summon_arg is None:
            return {"reason": "主机的 Summon 未初始化", "success": False}, 500

        new_game = GameSession(
            summon=summon_arg,
            mode=host_game.mode,
            drop_r=host_game.drop_r,
            ultimate_mode=host_game.ultimate_mode,
        )
        new_game.answer_board = host_game.answer_board.clone()
        new_game.board = host_game.board.clone()

        self.game = new_game
        self.board = new_game.board.clone()

        self.reset()

        return {"reason": "", "success": True}, 200



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/27 14:42
# @Author  : mojimoon
# @FileName: 1K2F.py
"""
[1K2F]马步花田: 染色格中的雷的八个马步位置内恰好有1个雷
"""

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard

class Rule1K2F(AbstractMinesRule):
    name = ["1K2F", "马步花田", "KnightFlower"]
    doc = "染色格中的雷的八个马步位置内恰好有1个雷"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, dye in board(mode="dye"):
            if not dye:
                continue
            positions = [_pos for _pos in pos.neighbors(5, 5)]
            # sum(vals) 表示马步格内的雷数
            sum_vals = sum(board.batch(positions, mode="variable", drop_none=True))
            val = board.get_variable(pos)
            # val 为1时，vals中必须有且仅有一个1
            # 约束：val=1 => sum(vals) == 1
            model.Add(sum_vals == 1).OnlyEnforceIf([val, s])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/10 11:07
# @Author  : xxx
# @FileName: 1L.py
"""
[1L]误差：所有线索均比真实值大1或小1
"""

from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition

from ....utils.tool import get_logger, get_random
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG


class Rule1L(AbstractClueRule):
    name = ["1L", "L", "误差"]
    doc = "所有线索均比真实值大1或小1"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        logger = get_logger()
        for pos, _ in board("N"):
            value = len([_pos for _pos in pos.neighbors(2) if board.get_type(_pos) == "F"])
            if value == 0:
                value = 1
            elif value == 8:
                value = 7
            else:
                value += 1 if random.random() > 0.5 else -1
            board.set_value(pos, Value1L(pos, count=value))
            logger.debug(f"Set {pos} to 1L[{value}]")
        return board


class Value1L(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            # 从字节码解码
            self.count = code[0]
        else:
            # 直接初始化
            self.count = count
        self.neighbor = self.pos.neighbors(2)
        self.pos = pos

    def __repr__(self):
        return f"{self.count}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbor

    @classmethod
    def type(cls) -> bytes:
        return Rule1L.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.count])

    def deduce_cells(self, board: 'AbstractBoard') -> bool:
        type_dict = {"N": [], "F": []}
        for pos in self.neighbor:
            t = board.get_type(pos)
            if t in ("", "C"):
                continue
            type_dict[t].append(pos)
        n_num, f_num = len(type_dict["N"]), len(type_dict["F"])
        if n_num == 0:
            return False
        if f_num == self.count + 1:
            for i in type_dict["N"]:
                board.set_value(i, VALUE_QUESS)
            return True
        if n_num + f_num == self.count - 1:
            for i in type_dict["N"]:
                board.set_value(i, MINES_TAG)
            return True
        return False

    def create_constraints(self, board: 'AbstractBoard', switch):
        """创建CP-SAT约束：周围雷数等于count"""
        model = board.get_model()
        s = switch.get(model, self)

        # 收集周围格子的布尔变量
        neighbor_vars = []
        for neighbor in self.neighbor:  # 8方向相邻格子
            if board.in_bounds(neighbor):
                var = board.get_variable(neighbor)
                neighbor_vars.append(var)

        # 添加约束：周围雷数等于count+-1
        if neighbor_vars:
            neighbor_sum = sum(neighbor_vars)
            # 两个布尔变量表示加和为 count + 1 或 count - 1
            b1 = model.NewBoolVar("sum_eq_count_plus_1")
            b2 = model.NewBoolVar("sum_eq_count_minus_1")

            # 将布尔变量与表达式绑定
            model.Add(neighbor_sum == self.count + 1).OnlyEnforceIf(b1)
            model.Add(neighbor_sum != self.count + 1).OnlyEnforceIf(b1.Not())

            model.Add(neighbor_sum == self.count - 1).OnlyEnforceIf(b2)
            model.Add(neighbor_sum != self.count - 1).OnlyEnforceIf(b2.Not())

            model.AddBoolOr([b1, b2]).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2025/06/07 13:45
# @Author  : Wu_RH
# @FileName: impl_obj.py
import base64
import os
import sys
import importlib.util
from pathlib import Path
from typing import Optional

from minesweepervariants.utils.tool import get_logger

from ..utils.impl_obj import VALUE_QUESS, MINES_TAG

from ..abs.rule import AbstractValue, AbstractRule
from ..abs.board import AbstractBoard
from ..abs.Lrule import AbstractMinesRule
from ..abs.Mrule import AbstractMinesClueRule, AbstractMinesValue
from ..abs.Rrule import AbstractClueRule, AbstractClueValue

from .board import version1, version2, version3
from . import rule

TOTAL = -1
hypothesis_board = [version3, version2, version1]


class ModelGenerateError(Exception):
    """模型求解器错误"""


def recursive_import(module):
    base_path = Path(module.__file__).parent
    base_name = module.__name__

    for dirpath, _, filenames in os.walk(base_path):
        for f in filenames:
            if f.endswith('.py') and f != '__init__.py':
                full_path = Path(dirpath) / f
                rel = full_path.relative_to(base_path).with_suffix('')
                mod_name = base_name + '.' + '.'.join(rel.parts)

                # 如果模块已经加载，跳过
                if mod_name in sys.modules:
                    continue

                # 尝试动态导入模块
                try:
                    spec = importlib.util.spec_from_file_location(str(mod_name), str(full_path))
                    if not spec or not spec.loader:
                        continue  # 跳过无效的 spec

                    mod = importlib.util.module_from_spec(spec)
                    sys.modules[mod_name] = mod  # 先注册到 sys.modules，避免循环导入问题
                    spec.loader.exec_module(mod)  # 执行模块代码
                except Exception as e:
                    get_logger().error(f"Failed to import {mod_name}: {e}")  # 打印错误信息（可选）
                    continue  # 跳过失败的模块


def set_total(total: int):
    global TOTAL
    TOTAL = total


def get_all_subclasses(cls):
    subclasses = set()
    direct_subs = cls.__subclasses__()
    subclasses.update(direct_subs)
    for sub in direct_subs:
        subclasses.update(get_all_subclasses(sub))
    return subclasses


def get_board(name: Optional[str] = None):
    if name is None:
        v = -1
        b = None
        for i in AbstractBoard.__subclasses__():
            if v < i.version:
                v = i.version
                b = i
        if b is None:
            raise ValueError("未找到棋盘")
        return b
    else:
        for i in AbstractBoard.__subclasses__():
            if i.name == name:
                return i


def get_rule(name: str) -> type | None:
    for i in get_all_subclasses(AbstractRule):
        if i in [
            AbstractClueRule,
            AbstractMinesClueRule,
            AbstractMinesRule
        ]:
            continue
        if type(i.name) in (tuple, list):
            if name in i.name:
                return i
            if name.upper() in i.name:
                return i
        elif type(i.name) is str:
            if name == i.name:
                return i
            if name.upper() == i.name:
                return i
    raise ValueError(f"未找到规则[{name}]")


def get_value(pos, code):
    code = code.split(b"|", 1)
    if code[0] == b"?":
        return VALUE_QUESS
    if code[0] == b"F":
        return MINES_TAG
    for i in get_all_subclasses(AbstractValue):
        if i in [
            AbstractValue,
            AbstractClueValue,
            AbstractMinesValue
        ]:
            continue
        if i.type() == code[0]:
            return i(pos=pos, code=code[1])
    return None


def encode_board(code: bytes) -> str:
    code = code[:]
    padding = len(code) % 4
    if padding:
        code += b'\xff' * (4 - padding)
    return base64.urlsafe_b64encode(code).decode("ascii")


def decode_board(base64data: str, name: Optional[str] = None):
    board_bytes = base64.urlsafe_b64decode(base64data.encode("ascii"))
    board_bytes = board_bytes.rstrip(b"\xff")
    return get_board(name)(code=board_bytes)


for pkg in [rule] + hypothesis_board:
    recursive_import(pkg)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 14:43
# @Author  : Wu_RH
# @FileName: 1S.py
"""
[3D1S] 三维蛇 (Snake)：所有雷构成一条蛇。蛇是一条宽度为 1 的六连通路径，不存在分叉、环、交叉
"""
from .. import Abstract3DMinesRule

from .connect import connect


class Rule1S(Abstract3DMinesRule):
    name = ["3D1S", "三维蛇", "3D-Snake"]
    doc = "所有雷构成一条蛇。蛇是一条宽度为 1 的六连通路径，不存在分叉、环、交叉"

    def create_constraints(self, board, switch):
        model = board.get_model()
        s = switch.get(model, self)

        connect(
            model=model,
            board=board,
            connect_value=1,
            nei_value=1,
            switch=s
        )

        tmp_list = []
        for pos, var in board(mode="variable"):
            tmp_bool = model.NewBoolVar("tmp")
            var_list = board.batch(self.pos_neighbors(board, pos, 1), mode="variable", drop_none=True)
            model.Add(sum(var_list) < 3).OnlyEnforceIf([var, s])
            model.Add(sum(var_list) == 1).OnlyEnforceIf([tmp_bool, s])
            model.Add(var == 1).OnlyEnforceIf([tmp_bool, s])
            tmp_list.append(tmp_bool)
        model.Add(sum(tmp_list) == 2).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/06/04 07:14
# @Author  : Wu_RH
# @FileName: impl_obj.py

from ..utils import tool

from ..abs.board import PositionTag
from ..abs.Mrule import MinesTag, ValueCircle
from ..abs.Rrule import ValueQuess, ValueCross

POSITION_TAG = PositionTag()
MINES_TAG = MinesTag(POSITION_TAG)
VALUE_QUESS = ValueQuess(POSITION_TAG)
VALUE_CIRCLE = ValueCircle(POSITION_TAG)
VALUE_CROSS = ValueCross(POSITION_TAG)

TOTAL = -1


def set_total(total: int):
    global TOTAL
    TOTAL = total


def get_total() -> int:
    return TOTAL


def get_seed():
    return tool.SEED



"""
[DC] 必对角 (Diagonally Connected)：每个雷对角相邻的四格中有至少一个雷
"""

from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.impl.summon.solver import Switch

class RuleDC(AbstractMinesRule):
    name = ["DC", "必对角", "Diagonally Connected"]
    doc = "每个雷对角相邻的四格中有至少一个雷"

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, var in board(mode="var"):
            var_list = board.batch(pos.neighbors(2, 2), mode="var", drop_none=True)
            model.Add(sum(var_list) >= 1).OnlyEnforceIf([var, s])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 19:14
# @Author  : Wu_RH
# @FileName: 1X'.py
"""
[1X'] 小十字 (Mini Cross)：线索表示半径为 1 的十字范围内的雷数
"""
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition

from ....utils.tool import get_logger
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG


class Rule1Xp(AbstractClueRule):
    name = ["1X'", "X'", "小十字", "Mini Cross"]
    doc = "线索表示半径为 1 的十字范围内的雷数"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        logger = get_logger()
        for pos, _ in board("N"):
            value = len([_pos for _pos in pos.neighbors(1) if board.get_type(_pos) == "F"])
            board.set_value(pos, Value1X(pos, count=value))
            logger.debug(f"Set {pos} to 1X[{value}]")
        return board


class Value1X(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            # 从字节码解码
            self.count = code[0]
        else:
            # 直接初始化
            self.count = count
        self.neighbor = self.pos.neighbors(1)

    def __repr__(self):
        return f"{self.count}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbor

    @classmethod
    def type(cls) -> bytes:
        return Rule1Xp.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.count])

    def deduce_cells(self, board: 'AbstractBoard') -> bool:
        type_dict = {"N": [], "F": []}
        for pos in self.neighbor:
            t = board.get_type(pos)
            if t in ("", "C"):
                continue
            type_dict[t].append(pos)
        n_num = len(type_dict["N"])
        f_num = len(type_dict["F"])
        if n_num == 0:
            return False
        if f_num == self.count:
            for i in type_dict["N"]:
                board.set_value(i, VALUE_QUESS)
            return True
        if f_num + n_num == self.count:
            for i in type_dict["N"]:
                board.set_value(i, MINES_TAG)
            return True
        return False

    def create_constraints(self, board: 'AbstractBoard', switch):
        """创建CP-SAT约束：周围雷数等于count"""
        model = board.get_model()
        s = switch.get(model, self)

        # 收集周围格子的布尔变量
        neighbor_vars = []
        for neighbor in self.neighbor:  # 8方向相邻格子
            if board.in_bounds(neighbor):
                var = board.get_variable(neighbor)
                neighbor_vars.append(var)

        # 添加约束：周围雷数等于count
        if neighbor_vars:
            model.Add(sum(neighbor_vars) == self.count).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/03 18:38
# @Author  : Wu_RH
# @FileName: Mrule.py

# 雷线索由于未实装 等待版本大更新

from typing import TYPE_CHECKING, Dict
from .rule import AbstractRule, AbstractValue
from ..utils.image_create import get_image, get_col, get_dummy, get_text
from abc import abstractmethod, ABC

from ..utils.web_template import Number

if TYPE_CHECKING:
    from minesweepervariants.abs.board import AbstractPosition, AbstractBoard


class AbstractMinesClueRule(AbstractRule, ABC):
    """
    雷线索规则
    """

    @abstractmethod
    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        """
        将在左线放置完成后调用
        需要将题板内的所有雷值赋值为线索
        :param board: 题板
        :return: 题板
        """


class AbstractMinesValue(AbstractValue, ABC):
    pos: 'AbstractPosition'

    @abstractmethod
    def __init__(self, pos: 'AbstractPosition', code: bytes = None):
        self.pos = pos

    def __repr__(self):
        """
        当前值在展示时候的显示字符串
        :return: 显示的字符串
        """
        return "F"

    def compose(self, board: 'AbstractBoard') -> Dict:
        """
        返回一个可渲染对象列表
        默认使用__repr__
        """
        return get_col(
            get_dummy(height=0.175),
            get_text(self.__repr__(),
                     color=("#FFFF00", "#FF7F00")),
            get_dummy(height=0.175),
        )

    def web_component(self, board: 'AbstractBoard') -> Dict:
        """
        返回一个可渲染对象列表
        默认使用__repr__
        """
        if "compose" in type(self).__dict__:
            return self.compose(board)
        data = Number(self.__repr__())
        return data


# --------实例类-------- #


class MinesTag(AbstractMinesValue):
    """
    雷标志类
    用于暂存表示为类
    """

    def __init__(self, pos: 'AbstractPosition', code: bytes = None):
        pass

    def __repr__(self):
        return "雷"

    def compose(self, board) -> Dict:
        return get_image(
            "flag",
            cover_pos_label=False
        )

    @classmethod
    def type(cls) -> bytes:
        return b"F"

    def code(self) -> bytes:
        return b""


class Rule0F(AbstractMinesClueRule):
    name = "_0F"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        self.drop = data is None

    def init_clear(self, board: 'AbstractBoard'):
        if not self.drop:
            return
        for key in board.get_board_keys():
            if not board.get_config(key, "interactive"):
                continue
            for pos, _ in board("F", key=key):
                board.set_value(pos, None)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        return board


class ValueCircle(AbstractMinesValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = None):
        pass

    def __repr__(self):
        return "O"

    def web_component(self, board: 'AbstractBoard') -> Dict:
        return get_image("circle", cover_pos_label=False)

    def compose(self, board) -> Dict:
        return get_image("circle", cover_pos_label=False)

    def code(self) -> bytes:
        return b""

    @classmethod
    def type(cls) -> bytes:
        return b"O"



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/28 03:49
# @Author  : Wu_RH
# @FileName: 3T.py
"""
[3T]无三连:任意三个雷不能等距排布
"""

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule1T(AbstractMinesRule):
    name = ["3T", "无三联"]
    doc = "任意三个雷不能等距排布"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        for key in board.get_interactive_keys():
            pos_bound = board.boundary(key=key)
            max_num = max(pos_bound.x, pos_bound.y) + 1
            for pos, _ in board():
                positions = []
                for i in range(-(max_num // 2 + 1), max_num // 2 + 1):
                    for j in range(max_num // 2 + 1):
                        if i == 0 and j == 0:
                            continue
                        positions.append([
                            pos, pos.shift(i, j),
                            pos.shift(2*i, 2*j)
                        ])
                for position in positions:
                    var_list = board.batch(position, mode="variable")
                    if True in [i is None for i in var_list]:
                        continue
                    model.Add(sum(var_list) != 3).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/10 11:07
# @Author  : xxx
# @FileName: 1L.py
"""
[2M]取模:线索与周围8格的雷数除以3的余数相同
"""

from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition

from ....utils.tool import get_logger


class Rule2M(AbstractClueRule):
    name = ["2M", "取模", "Mod"]
    doc = "线索与周围8格的雷数除以3的余数相同"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        logger = get_logger()
        for pos, _ in board("N"):
            value = len([_pos for _pos in pos.neighbors(2)
                         if board.get_type(_pos) == "F"]) % 3
            board.set_value(pos, Value2M(pos, count=value))
            logger.debug(f"Set {pos} to 2M[{value}]")
        return board


class Value2M(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            # 从字节码解码
            self.count = code[0]
        else:
            # 直接初始化
            self.count = count
        self.neighbor = self.pos.neighbors(2)

    def __repr__(self):
        return f"{self.count}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbor

    @classmethod
    def type(cls) -> bytes:
        return Rule2M.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.count])

    def create_constraints(self, board: 'AbstractBoard', switch):
        """创建CP-SAT约束：周围雷数等于count"""
        model = board.get_model()
        s = switch.get(model, self)

        # 收集周围格子的布尔变量
        neighbor_vars = []
        for neighbor in self.neighbor:  # 8方向相邻格子
            if board.in_bounds(neighbor):
                var = board.get_variable(neighbor)
                neighbor_vars.append(var)

        # 添加约束：周围雷数等于count+-1
        if neighbor_vars:
            neighbor_sum = sum(neighbor_vars)

            b1 = model.NewBoolVar(f"[2M]b1")
            b2 = model.NewBoolVar(f"[2M]b2")
            b3 = model.NewBoolVar(f"[2M]b3")

            # 将布尔变量与表达式绑定
            model.Add(neighbor_sum == self.count).OnlyEnforceIf([b1, s])
            model.Add(neighbor_sum != self.count).OnlyEnforceIf([b1.Not(), s])

            model.Add(neighbor_sum == self.count + 3).OnlyEnforceIf([b2, s])
            model.Add(neighbor_sum != self.count + 3).OnlyEnforceIf([b2.Not(), s])

            model.Add(neighbor_sum == self.count + 6).OnlyEnforceIf([b3, s])
            model.Add(neighbor_sum != self.count + 6).OnlyEnforceIf([b3.Not(), s])

            model.AddBoolOr([b1, b2, b3]).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2025/06/03 04:23
# @Author  : Wu_RH
# @FileName: __init__.py

import os
import ast
from typing import Dict, Union


def extract_module_docstring(filepath) -> Union[Dict, None]:
    if "sharp" in filepath:
        print(filepath)
    try:
        with open(filepath, 'r', encoding='utf-8') as f:
            source = f.read()
    except Exception:
        return None

    try:
        tree = ast.parse(source, filename=filepath)
    except SyntaxError:
        return None

    module_doc = ast.get_docstring(tree)

    for node in ast.walk(tree):
        if not isinstance(node, ast.ClassDef):
            continue

        bases_info = []
        for base in node.bases:
            if isinstance(base, ast.Name):
                bases_info.append(base.id)
            elif isinstance(base, ast.Attribute):
                bases_info.append(base.attr)
            else:
                bases_info.append(str(base))

        x = 0
        if any("Mines" in b for b in bases_info):
            x |= 1
        if any("MinesClue" in b for b in bases_info):
            x |= 2
        if any("Clue" in b for b in bases_info):
            x |= 4

        if x == 6:
            x = 2

        info = {}
        for stmt in node.body:
            if isinstance(stmt, ast.Assign):
                for target in stmt.targets:
                    if (
                        isinstance(target, ast.Name) and target.id == "name" and
                        (isinstance(stmt.value, ast.Str) or
                         isinstance(stmt.value, ast.List) or
                         isinstance(stmt.value, ast.Tuple))
                    ):
                        # 处理字符串列表的情况
                        if (
                            isinstance(stmt.value, ast.List) or
                            isinstance(stmt.value, ast.Tuple)
                        ):
                            # 提取列表中的第一个字符串作为名称
                            name_vals = []
                            for elt in stmt.value.elts:
                                if isinstance(elt, ast.Str) and elt.s.strip():
                                    name_val = elt.s.strip()
                                    name_vals.append(name_val)
                            info["x"] = x
                            info["module_doc"] = module_doc
                            info["names"] = name_vals
                        # 处理单个字符串的情况
                        elif isinstance(stmt.value, ast.Str) and stmt.value.s.strip():
                            name_val = stmt.value.s.strip()
                            info["x"] = x
                            info["module_doc"] = module_doc
                            info["names"] = [name_val]
                    if (
                        isinstance(target, ast.Name) and
                        target.id == "doc" and
                        isinstance(stmt.value, ast.Str)
                    ):
                        # 处理单个字符串的情况
                        doc_val = stmt.value.s.strip()
                        info["doc"] = doc_val
        if "names" in info:
            return info
    return None


def scan_module_docstrings(directory):
    results = []
    for root, _, files in os.walk(directory):
        for name in files:
            if name.endswith('.py'):
                path = os.path.join(root, name)
                pck = extract_module_docstring(path)
                if pck is None:
                    continue
                m_doc = pck.get('module_doc', "")
                x = pck.get('x', 0)
                names = pck.get('names', [])
                doc = pck.get('doc', "")
                results.append((m_doc, doc, x, names))
    return results


def get_all_rules():
    results = {"R": {}, "M": {}, "L": {}, "O": {}}
    dir_path = os.path.dirname(os.path.abspath(__file__))
    for m_doc, doc, x, names in scan_module_docstrings(dir_path):
        if not names:
            continue
        name, names = names[0], names[1:]
        rule_line = None
        if x == 1:
            rule_line = "L"
        elif x == 2:
            rule_line = "M"
        elif x == 4:
            rule_line = "R"
        if rule_line is None:
            continue
        results[rule_line][name] = {
            "names": names,
            "doc": doc,
            "module_doc": m_doc,
        }
    return results



#!/usr/bin/env python3
"""
[3N] 范数 (Norm)：线索(a_p)表示距离自己lp范数最近的雷的lp范数大小为a。(p=0,1,2,00)
"""
from typing import Dict, Literal

from ....abs.Rrule import AbstractClueValue, AbstractClueRule
from ....abs.board import AbstractPosition, AbstractBoard
from ....utils.image_create import get_text, get_image, get_row, get_col, get_dummy
from ....utils.tool import get_random

P = Literal[0, 1, 2] | Literal['00']
Root = tuple[int, int]


def simplify_sqrt(n):
    c = 1
    i = 2
    while i * i <= n:
        while n % (i * i) == 0:
            c *= i
            n //= i * i
        i += 1
    return c, n


def norm(dx, dy, p: P) -> Root:
    """
    计算点(dx, dy)在lp范数下的距离

    参数:
        dx: x方向的距离
        dy: y方向的距离
        p: lp范数的阶数

    返回:
        点(dx, dy)在lp范数下的距离
    """
    if p == 0:
        return 2 if dx != 0 and dy != 0 else 1, 1
    elif p == 1:
        return abs(dx) + abs(dy), 1
    elif p == 2:
        r2 = dx ** 2 + dy ** 2
        if int(r2 ** 0.5) ** 2 == r2:
            return int(r2 ** 0.5), 1
        else:
            return r2, 2
    elif p == '00':
        return max(abs(dx), abs(dy)), 1
    else:
        raise ValueError(f"Unsupported norm type: {p}")


def format(n: Root, p: P):
    """
    格式化输出范数的字符串表示
    """
    if n[1] == 1:
        return f'{n[0]}_{p}'
    elif n[1] == 2:
        c, n_ = simplify_sqrt(n[0])
        return f"{c if c > 1 else ''}√{n_}" if n_ > 1 else str(c) + f'_{p}'
    else:
        raise ValueError(f"Unsupported root type: {n[1]}")


class BaseRule3N(AbstractClueRule):
    name = ["3N", "范数", "Norm"]
    doc = "线索(a_p)表示距离自己lp范数最近的雷的lp范数大小为a。(p=0,1,2,00)"
    p: P = -1

    def __init__(self, board: AbstractBoard, data: list[AbstractClueRule] = None):
        super().__init__(None, None)
        for key in board.get_interactive_keys():
            board.set_config(key, "by_mini", True)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        boards = []
        for rule in [Rule3N0(board), Rule3N1(board),
                     Rule3N2(board), Rule3NInf(board)]:
            boards.append(rule._fill(board.clone()))
        for key in board.get_board_keys():
            for pos, _ in board("N", key=key):
                values = [_board.get_value(pos)
                          for _board in boards
                          if _board.get_type(pos) != "N"]
                if not values:
                    continue
                board.set_value(pos, get_random().choice(values))
        return board

    #
    def _fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        if len([_ for _ in board("F")]) < 1:
            return board

        for pos, _ in board("N"):
            p: P = self.p
            min_norm = None

            # 遍历所有雷的位置
            for mine_pos, _ in board("F"):
                dx = mine_pos.x - pos.x
                dy = mine_pos.y - pos.y
                mine_norm = norm(dx, dy, p)

                if min_norm is None or (mine_norm[0] < min_norm[0] or
                                        (mine_norm[0] == min_norm[0] and mine_norm[1] < min_norm[1])):
                    min_norm = mine_norm

            if min_norm is not None:
                # 编码范数值和p值
                if p == '00':
                    p_val = 100
                else:
                    p_val = int(p)
                obj = self.clue_class()(pos, bytes([min_norm[0], min_norm[1], p_val]))
                board.set_value(pos, obj)

        return board

    def clue_class(self):
        return getattr(self, 'clue', BaseValue3N)


class BaseValue3N(AbstractClueValue):
    name = "3N"

    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.pos = pos
        if len(code) >= 3:
            self.n = (code[0], code[1])  # (数值, 根次数)
            if code[2] == 100:
                self.p: P = '00'
            elif code[2] == 0:
                self.p: P = 0
            elif code[2] == 1:
                self.p: P = 1
            elif code[2] == 2:
                self.p: P = 2
            else:
                self.p: P = 2  # 默认值
        else:
            self.n = (1, 1)
            self.p: P = 2

    def __repr__(self):
        return format(self.n, self.p)

    def web_component(self, board) -> Dict:
        if self.n[1] == 1:
            return get_text(str(self.n[0]))
        elif self.n[1] == 2:
            value_a, value_b = simplify_sqrt(self.n[0])
            if value_b == -1:
                return get_text(str(value_a))
            if value_a == -1:
                return get_text(
                    "$\\sqrt{" + str(value_b) + "}$"
                )
            else:
                return get_text(
                    "$" + str(value_a) +
                    "\\sqrt{" + str(value_b) +
                    "}$"
                )
        else:
            raise ValueError("Unsupported root type")

    def compose(self, board) -> Dict:
        if self.n[1] == 1:
            return get_col(
                get_dummy(height=0.175),
                get_text(str(self.n[0])),
                get_dummy(height=0.175),
            )
        elif self.n[1] == 2:
            value_a, value_b = simplify_sqrt(self.n[0])
            if value_a == 1:
                return get_row(
                    get_image("sqrt"),
                    get_text(str(value_b)),
                    spacing=-0.15
                )
            else:
                return get_row(
                    get_text(str(value_a)),
                    get_image("sqrt"),
                    get_text(str(value_b)),
                    spacing=-0.2
                )
        else:
            raise ValueError("Unsupported root type")

    @classmethod
    def type(cls) -> bytes:
        return cls.name.encode('ascii')

    def code(self) -> bytes:
        if self.p == '00':
            p_val = 100
        else:
            p_val = int(self.p)
        return bytes([self.n[0], self.n[1], p_val])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        # 为当前线索位置创建约束
        # 确保存在一个雷，其lp范数等于线索值
        constraint_vars = []

        for mine_pos, mine_var in board(mode="variable"):
            if board.get_type(mine_pos) in ["N", "F"]:  # 可能的雷位置
                dx = mine_pos.x - self.pos.x
                dy = mine_pos.y - self.pos.y

                # 计算这个位置到线索位置的lp范数
                pos_norm = norm(dx, dy, self.p)

                # 如果范数匹配，则这个位置可能是最近的雷
                if pos_norm == self.n:
                    # 创建一个布尔变量表示这个雷是最近的
                    is_closest = model.NewBoolVar(f"[3N]_{self.pos}_{mine_pos}")
                    model.Add(mine_var == 1).OnlyEnforceIf([is_closest, s])

                    # 确保没有更近的雷
                    for other_pos, other_var in board(mode="variable"):
                        if other_pos != mine_pos and board.get_type(other_pos) in ["N", "F"]:
                            other_dx = other_pos.x - self.pos.x
                            other_dy = other_pos.y - self.pos.y
                            other_norm = norm(other_dx, other_dy, self.p)

                            # 如果有更近的雷，则当前雷不能是最近的
                            if (other_norm[0] < pos_norm[0] or
                                    (other_norm[0] == pos_norm[0] and other_norm[1] < pos_norm[1])):
                                model.Add(other_var == 0).OnlyEnforceIf([is_closest, s])

                    constraint_vars.append(is_closest)

        # 至少有一个雷满足条件
        if constraint_vars:
            model.AddBoolOr(constraint_vars).OnlyEnforceIf(s)


class Value3N0(BaseValue3N):
    name = "3N0"
    p = 0


class Rule3N0(BaseRule3N):
    name = "3N0"
    p = 0
    clue = Value3N0


class Value3N1(BaseValue3N):
    name = "3N1"
    p = 1


class Rule3N1(BaseRule3N):
    name = "3N1"
    p = 1
    clue = Value3N1


class Value3N2(BaseValue3N):
    name = "3N2"
    p = 2


class Rule3N2(BaseRule3N):
    name = "3N2"
    p = 2
    clue = Value3N2


class Value3NInf(BaseValue3N):
    name = "3N00"
    p = '00'


class Rule3NInf(BaseRule3N):
    name = "3N00"
    p = '00'
    clue = Value3NInf



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/10 09:52
# @Author  : xxx
# @FileName: c.py

from . import AbstractDye


class DyeC(AbstractDye):
    name = "q" # quadrant
    __doc__ = "2x2棋盘格染色"

    def parse_int(self, str_value, default):
        if str_value is None:
            return default
        try:
            return int(str_value)
        except ValueError:
            return default

    def __init__(self, args):
        '''
        args: size:dx:dy
        size: 棋盘格大小，默认2
        dx, dy: 棋盘格偏移，默认0
        '''
        if not args:
            self.size = 2
            self.dx = 0
            self.dy = 0
        else:
            all_args = args.split(':')
            self.size = self.parse_int(all_args[0], 2)
            self.dx = self.parse_int(all_args[1], 0) if len(all_args) > 1 else 0
            self.dy = self.parse_int(all_args[2], 0) if len(all_args) > 2 else 0

    def dye(self, board):
        dye = True
        for key in board.get_interactive_keys():
            dye = not dye
            for pos, _ in board(key=key):
                # if self.offset == 0:
                #     _dye = dye ^ ((pos.x // 2 + pos.y // 2) % 2 > 0)
                # elif self.offset == 1:
                #     _dye = dye ^ (((pos.x + 1) // 2 + pos.y // 2) % 2 > 0)
                # elif self.offset == 2:
                #     _dye = dye ^ ((pos.x // 2 + (pos.y + 1) // 2) % 2 > 0)
                # else:
                #     _dye = dye ^ (((pos.x + 1)//2 + (pos.y + 1)//2) % 2 > 0)
                _dye = dye ^ (((pos.x + self.dx) // self.size + (pos.y + self.dy) // self.size) % 2 > 0)
                board.set_dyed(pos, _dye)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 18:26
# @Author  : Wu_RH
# @FileName: 1X.py
"""
[1X] 十字 (Cross)：线索表示半径为 2 的十字范围内的雷数
"""
from typing import List

from minesweepervariants.impl.summon.solver import Switch
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition

from ....utils.impl_obj import VALUE_QUESS, MINES_TAG


def encode_bools_7bit(bools: list[bool]) -> bytes:
    """
    将布尔列表转换为字节，每7个布尔值转换为1个字节
    参数:
        bool_list: 布尔值列表
    返回:
        bytes: 转换后的字节对象
    """
    if len(bools) == 0:
        return b''
    bools = [False] * (7 - len(bools) % 7) + bools
    byte_array = bytearray()

    # 每7个布尔值处理为一个字节
    for i in range(0, len(bools), 7):
        group = bools[i:i + 7]
        byte_val = 0

        # 将7个布尔值转换为一个字节
        for j, bit in enumerate(group):
            if bit:
                byte_val |= 1 << (6 - j)  # 设置相应的位

        byte_array.append(byte_val)

    return bytes(byte_array)


def decode_bools_7bit(data: bytes) -> list[bool]:
    # 解码所有数据，每个字节转换为7个布尔值
    bools = []
    for byte in data:
        for shift in range(6, -1, -1):
            bit = (byte >> shift) & 1
            bools.append(bool(bit))

    return bools


class Rule1X(AbstractClueRule):
    name = ["1X", "十字", "Cross"]
    doc = "线索表示半径为 2 的十字范围内的雷数"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        if data is None:
            self.neibor_bool = []
            return
        datas = data.split(";")
        nei_values: list[int] = []
        for nei_value in datas:
            if ":" in nei_value:
                nei_values.extend(tuple([
                    i for i in range(
                        int(nei_value.split(":")[0]),
                        int(nei_value.split(":")[1])+1
                    )
                ]))
            else:
                nei_values.append(int(nei_value))
        if not nei_values:
            return
        max_value = max(nei_values)
        self.neibor_bool = [False for _ in range(max_value)]
        for i in nei_values:
            if i == 0:
                continue
            self.neibor_bool[max_value-i] = True

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        code = b'\x00' + encode_bools_7bit(self.neibor_bool)
        for pos, _ in board("N"):
            obj = Value1X(pos, code)
            value = len([_pos for _pos in obj.neighbor if board.get_type(_pos) == "F"])
            obj.value = value
            board.set_value(pos, obj)
        return board


class Value1X(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, code: bytes = None):
        super().__init__(pos, code)
        self.value, self.data = code[0], code[1:]
        if self.data == b'':
            bool_list = [True, False, False, True]
        else:
            data = self.data
            bool_list = decode_bools_7bit(data)
        self.neighbor = []
        print(bool_list)
        for n, b in enumerate(bool_list[::-1]):
            if not b:
                continue
            self.neighbor.extend(self.pos.neighbors(n + 1, n + 1))

    def __repr__(self):
        return f"{self.value}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbor

    @classmethod
    def type(cls) -> bytes:
        return Rule1X.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value]) + self.data

    def deduce_cells(self, board: 'AbstractBoard') -> bool:
        type_dict = {"N": [], "F": []}
        for pos in self.neighbor:
            t = board.get_type(pos)
            if t in ("", "C"):
                continue
            type_dict[t].append(pos)
        n_num = len(type_dict["N"])
        f_num = len(type_dict["F"])
        if n_num == 0:
            return False
        if f_num == self.value:
            for i in type_dict["N"]:
                board.set_value(i, VALUE_QUESS)
            return True
        if f_num + n_num == self.value:
            for i in type_dict["N"]:
                board.set_value(i, MINES_TAG)
            return True
        return False

    def create_constraints(self, board: 'AbstractBoard', switch):
        """创建CP-SAT约束：周围雷数等于count"""
        model = board.get_model()
        s = switch.get(model, self)

        # 收集周围格子的布尔变量
        neighbor_vars = []
        for neighbor in self.neighbor:  # 8方向相邻格子
            if board.in_bounds(neighbor):
                var = board.get_variable(neighbor)
                neighbor_vars.append(var)

        # 添加约束：周围雷数等于count
        if neighbor_vars:
            model.Add(sum(neighbor_vars) == self.value).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/12 19:40
# @Author  : Wu_RH
# @FileName: 2Z'.py
"""
[2Z']零和'(Zero-Sum'):每个4x4方块内的染色格和非染色格雷数相等
"""
from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard, AbstractPosition
from minesweepervariants.impl.summon.solver import Switch


def block4(board: AbstractBoard, pos: AbstractPosition):
    positions = []
    pos_list = [pos.right(i) for i in range(4)]
    for pos in pos_list:
        positions.extend([pos.down(i) for i in range(4)])
    if '' in board.batch(positions, mode="type"):
        return None
    return positions


class Rule2Zp(AbstractMinesRule):
    name = ["2Z'", "零和'", "Zero-Sum'"]
    doc = "每个4x4方块内的染色格和非染色格雷数相等"

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, _ in board():
            block = block4(board, pos)
            if block is None:
                continue
            dye_list = board.batch(block, mode="dye")
            var_list = board.batch(block, mode="var")
            var1_list = [v for v, d in zip(var_list, dye_list) if d]
            var2_list = [v for v, d in zip(var_list, dye_list) if not d]

            model.Add(sum(var1_list) == sum(var2_list)).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/15 09:01
# @Author  : Wu_RH
# @FileName: 3Q.py
"""
[3Q]正方形(Square): 所有四连通的雷格区域组成实心正方形
"""
from typing import List

from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard, AbstractPosition
from minesweepervariants.impl.summon.solver import Switch


def block(a_pos: AbstractPosition, board: AbstractBoard) -> List[AbstractPosition]:
    b_pos = a_pos.up()
    c_pos = a_pos.left()
    d_pos = b_pos.left()
    if not board.in_bounds(d_pos):
        return []
    return [a_pos, b_pos, c_pos, d_pos]


class Rule3Q(AbstractMinesRule):
    name = ["3Q", "正方形", "Square"]
    doc = "所有四连通的雷格区域组成实心正方形"

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)

        for pos, var in board(mode="var"):
            block_pos = block(pos, board)
            if block_pos:
                block_var = board.batch(block_pos, mode="var")
                model.Add(sum(block_var) != 3)

            row = board.get_row_pos(pos)
            col = board.get_col_pos(pos)

            var_list = [var]
            if board.in_bounds(pos.up()):
                var_list.append(board.get_variable(pos.up()).Not())
            if board.in_bounds(pos.left()):
                var_list.append(board.get_variable(pos.left()).Not())

            tmp_list = []

            for row_pos, col_pos in zip(
                row[row.index(pos):],
                col[col.index(pos):]
            ):
                tmp = model.NewBoolVar(f"tmp[{pos}, {row_pos}, {col_pos}]")
                model.AddBoolAnd(var_list).OnlyEnforceIf(tmp)
                row_box = board.get_pos_box(pos, row_pos)
                col_box = board.get_pos_box(pos, col_pos)
                row_var = board.batch(row_box, "var")
                col_var = board.batch(col_box, "var")
                model.AddBoolAnd(row_var).OnlyEnforceIf(tmp)
                model.AddBoolAnd(col_var).OnlyEnforceIf(tmp)
                if board.is_valid(row_pos.right()):
                    model.AddBoolAnd(board.get_variable(row_pos.right()).Not()).OnlyEnforceIf(tmp)
                if board.is_valid(col_pos.down()):
                    model.AddBoolAnd(board.get_variable(col_pos.down()).Not()).OnlyEnforceIf(tmp)
                tmp_list.append(tmp)
            model.AddBoolOr(tmp_list).OnlyEnforceIf(var_list + [s])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/09 01:30
# @Author  : Wu_RH
# @FileName: B1.py
"""
[1B]平衡: 每行每列雷数相同
"""
from math import gcd

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule1B(AbstractMinesRule):
    name = ["1B", "平衡"]
    doc = "1.每一列的雷数相同 且2.每一行的雷数相同"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s1 = switch.get(model, self)
        s2 = switch.get(model, self)
        for key in board.get_interactive_keys():
            boundary_pos = board.boundary(key=key)

            row_positions = board.get_row_pos(boundary_pos)
            row_sums = [
                sum(board.get_variable(_pos) for _pos in board.get_col_pos(pos))
                for pos in row_positions
            ]
            # 所有 row_sums 相等
            for i in range(1, len(row_sums)):
                model.Add(row_sums[i] == row_sums[0]).OnlyEnforceIf(s1)

            col_positions = board.get_col_pos(boundary_pos)
            col_sums = [
                sum(board.get_variable(_pos) for _pos in board.get_row_pos(pos))
                for pos in col_positions
            ]
            # 所有 col_sums 相等
            for i in range(1, len(col_sums)):
                model.Add(col_sums[i] == col_sums[0]).OnlyEnforceIf(s2)

    def suggest_total(self, info: dict):
        def lcm(a, b):
            return a * b // gcd(a, b) if a and b else 0

        def add_constraints(model, total_var):
            """实际添加约束的函数（保持最小化）"""
            nonlocal bases

            # 创建分量变量
            components = []
            for i, base in enumerate(bases):
                max_multiplier = (total_var.Proto().domain[1] // base) + 1
                mult = model.NewIntVar(0, max_multiplier, f"mult_{i}")
                product = model.NewIntVar(0, total_var.Proto().domain[1], f"prod_{i}")

                # 添加乘积约束
                model.AddMultiplicationEquality(product, [mult, base])
                components.append(product)

            # 添加总和约束
            model.Add(sum(components) == total_var)

        sizes = [info["size"][interactive] for interactive in info["interactive"]]

        # 预处理阶段：计算各size的base_i
        bases = [lcm(w, h) for w, h in sizes if w > 0 and h > 0]

        info["hard_fns"].append(add_constraints)



#!/usr/bin/env python3

"""
[CQ] 王后 (Chess-Queen)：线索表示八方向上雷数总和
"""
from typing import List

from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition

from ....utils.tool import get_logger
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG


def _get_queen_positions(board: 'AbstractBoard', pos: AbstractPosition):
    """获取与给定位置在王后移动范围内的所有位置（横纵+斜向）"""
    positions = []
    # 获取棋盘的边界
    boundary = board.boundary()
    max_x, max_y = boundary.x, boundary.y

    # 横向方向（同行，相同x，不同y）
    for y in range(max_y + 1):
        other_pos = type(pos)(pos.x, y, pos.board_key)
        if other_pos != pos and board.in_bounds(other_pos):
            positions.append(other_pos)

    # 纵向方向（同列，相同y，不同x）
    for x in range(max_x + 1):
        other_pos = type(pos)(x, pos.y, pos.board_key)
        if other_pos != pos and board.in_bounds(other_pos):
            positions.append(other_pos)

    # 右上斜线方向 (x+1, y+1)
    for i in range(1, max(max_x, max_y) + 1):
        other_pos = type(pos)(pos.x + i, pos.y + i, pos.board_key)
        if board.in_bounds(other_pos):
            positions.append(other_pos)
        else:
            break

    # 左下斜线方向 (x-1, y-1)
    for i in range(1, max(max_x, max_y) + 1):
        other_pos = type(pos)(pos.x - i, pos.y - i, pos.board_key)
        if board.in_bounds(other_pos):
            positions.append(other_pos)
        else:
            break

    # 左上斜线方向 (x-1, y+1)
    for i in range(1, max(max_x, max_y) + 1):
        other_pos = type(pos)(pos.x - i, pos.y + i, pos.board_key)
        if board.in_bounds(other_pos):
            positions.append(other_pos)
        else:
            break

    # 右下斜线方向 (x+1, y-1)
    for i in range(1, max(max_x, max_y) + 1):
        other_pos = type(pos)(pos.x + i, pos.y - i, pos.board_key)
        if board.in_bounds(other_pos):
            positions.append(other_pos)
        else:
            break

    return positions


class Rule1XStar(AbstractClueRule):
    name = ["CQ", "王后", "Chess-Queen"]
    doc = "线索表示八方向上雷数总和"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        logger = get_logger()
        for pos, _ in board("N"):
            # 计算斜向和横纵所有格子中的雷数
            queen_positions = _get_queen_positions(board, pos)
            value = len([_pos for _pos in queen_positions if board.get_type(_pos) == "F"])
            board.set_value(pos, Value1XStar(pos, count=value))
            logger.debug(f"Set {pos} to 1X*[{value}]")
        return board


class Value1XStar(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            # 从字节码解码
            self.count = code[0]
        else:
            # 直接初始化
            self.count = count

    def _get_queen_positions(self, board: 'AbstractBoard'):
        """获取与给定位置在王后移动范围内的所有位置（横纵+斜向）"""
        positions = []
        # 获取棋盘的边界
        boundary = board.boundary()
        max_x, max_y = boundary.x, boundary.y

        # 横向方向（同行，相同x，不同y）
        for y in range(max_y + 1):
            other_pos = type(self.pos)(self.pos.x, y, self.pos.board_key)
            if other_pos != self.pos and board.in_bounds(other_pos):
                positions.append(other_pos)

        # 纵向方向（同列，相同y，不同x）
        for x in range(max_x + 1):
            other_pos = type(self.pos)(x, self.pos.y, self.pos.board_key)
            if other_pos != self.pos and board.in_bounds(other_pos):
                positions.append(other_pos)

        # 右上斜线方向 (x+1, y+1)
        for i in range(1, max(max_x, max_y) + 1):
            other_pos = type(self.pos)(self.pos.x + i, self.pos.y + i, self.pos.board_key)
            if board.in_bounds(other_pos):
                positions.append(other_pos)
            else:
                break

        # 左下斜线方向 (x-1, y-1)
        for i in range(1, max(max_x, max_y) + 1):
            other_pos = type(self.pos)(self.pos.x - i, self.pos.y - i, self.pos.board_key)
            if board.in_bounds(other_pos):
                positions.append(other_pos)
            else:
                break

        # 左上斜线方向 (x-1, y+1)
        for i in range(1, max(max_x, max_y) + 1):
            other_pos = type(self.pos)(self.pos.x - i, self.pos.y + i, self.pos.board_key)
            if board.in_bounds(other_pos):
                positions.append(other_pos)
            else:
                break

        # 右下斜线方向 (x+1, y-1)
        for i in range(1, max(max_x, max_y) + 1):
            other_pos = type(self.pos)(self.pos.x + i, self.pos.y - i, self.pos.board_key)
            if board.in_bounds(other_pos):
                positions.append(other_pos)
            else:
                break

        return positions

    def __repr__(self):
        return f"{self.count}"

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition']:
        return self._get_queen_positions(board)

    @classmethod
    def type(cls) -> bytes:
        return Rule1XStar.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.count])

    def deduce_cells(self, board: 'AbstractBoard') -> bool:
        queen_positions = self._get_queen_positions(board)
        type_dict = {"N": [], "F": []}

        for pos in queen_positions:
            t = board.get_type(pos)
            if t in ("", "C"):
                continue
            type_dict[t].append(pos)

        n_num = len(type_dict["N"])
        f_num = len(type_dict["F"])

        if n_num == 0:
            return False

        # 如果已找到的雷数等于目标数，剩余格子都是安全的
        if f_num == self.count:
            for i in type_dict["N"]:
                board.set_value(i, VALUE_QUESS)
            return True

        # 如果已找到的雷数加上未知格子数等于目标数，剩余格子都是雷
        if f_num + n_num == self.count:
            for i in type_dict["N"]:
                board.set_value(i, MINES_TAG)
            return True

        return False

    def create_constraints(self, board: 'AbstractBoard', switch):
        """创建CP-SAT约束：王后移动范围内格子的雷数等于count"""
        model = board.get_model()
        s = switch.get(model, self)

        # 收集王后移动范围内格子的布尔变量
        queen_positions = self._get_queen_positions(board)
        neighbor_vars = []

        for neighbor in queen_positions:
            if board.in_bounds(neighbor):
                var = board.get_variable(neighbor)
                neighbor_vars.append(var)

        # 添加约束：王后移动范围内格子的雷数等于count
        if neighbor_vars:
            model.Add(sum(neighbor_vars) == self.count).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/06 02:11
# @Author  : QuirkyStorm7988
# @FileName: 1W.py
"""
[1W] 数墙 (Wall)：线索表示 3x3 范围内每组连续雷的长度
"""
from typing import Dict

from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractPosition, AbstractBoard
from ....utils.image_create import get_text, get_row, get_col
from ....utils.image_create import get_dummy
from ....utils.tool import get_logger

from ....utils.web_template import MultiNumber


def decode(code: bytes) -> list[int]:
    if len(code) == 2:
        if code[1] > 0xf:
            return [code[1] >> 4, code[1] & 0xf, code[0] >> 4, code[0] & 0xf]
        return [code[1] & 0xf, code[0] >> 4, code[0] & 0xf]
    elif len(code) == 1:
        if code[0] > 0xf:
            return [code[0] >> 4, code[0] & 0xf]
        return [code[0] & 0xf]
    return []


def encode(values: list[int]) -> bytes:
    if len(values) == 1:
        return bytes([values[0] & 0xf])

    if len(values) == 2:
        return bytes([(values[0] << 4) | (values[1] & 0xf)])

    if len(values) == 3:
        b0 = ((values[1] & 0xf) << 4) | (values[2] & 0xf)
        b1 = values[0] & 0xf
        return bytes([b0, b1])

    if len(values) == 4:
        b0 = ((values[2] & 0xf) << 4) | (values[3] & 0xf)
        b1 = ((values[0] & 0xf) << 4) | (values[1] & 0xf)
        return bytes([b0, b1])

    return b""


def MineStatus_1W(clue: list) -> list:
    """
    返回值：一个int列表，其中存的每一个int表示：
        一个二进制数，第i位（从低到高）表示从左上角开始顺时针旋转，第i个格子的雷情况（是雷->1，非雷->0）
        将这个二进制数转化为十进制存储到元素当中，如42(10) == 00101010(2)，即这个线索格的右上、右下、左下有雷
    """
    ans = []
    a = [0 for _ in range(8)]  # 决策列表

    def dfs(step: int):
        if step >= 8:  # 最终处理
            # 先写没有剪枝的
            test = []
            last = 0
            for i in range(8):
                if a[i]:
                    last += 1
                else:
                    if last != 0: test.append(last)
                    last = 0
            if last != 0: test.append(last)
            if a[-1] and a[0] and len(test) != 1:
                test[0] += test[-1]
                del test[-1]
            if not test: test = [0]
            test.sort()
            if test != clue: return None
            #
            status = 0
            for i in range(8):
                status += 2 ** i * a[i]
            if status not in ans:
                ans.append(status)
            # if a[:] not in ans:
            #     ans.append(a[:])
            return None
        a[step] = 0
        dfs(step + 1)
        a[step] = 1
        dfs(step + 1)
        return None

    dfs(0)
    return ans


class Rule1W(AbstractClueRule):
    name = ["1W", "Wall", "数墙"]
    doc = "线索表示 3x3 范围内每组连续雷的长度"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        logger = get_logger()
        for pos, _ in board("N"):
            nei = [pos.right(), pos.right().down(), pos.down(), pos.left().down(),
                   pos.left(), pos.left().up(), pos.up(), pos.right().up()]
            values = []
            nei_type = board.batch(nei, mode="type")
            value = 0
            t = ""
            for t in nei_type:
                if t == "F":
                    value += 1
                elif value != 0:
                    values.append(value)
                    value = 0
            if value != 0 and t == nei_type[0] == "F":
                values[0] += value
            elif value != 0:
                values.append(value)
            values.sort()
            obj = Value1W(pos, encode(values))
            board.set_value(pos, obj)
            logger.debug(f"[1W]set {obj} to {pos}")

        return board


class Value1W(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.values: list[int] = decode(code)
        self.pos = pos

    def __repr__(self):
        return ".".join([str(i) for i in self.values]) if len(self.values) > 0 else "0"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.pos.neighbors(2)

    def web_component(self, board) -> Dict:
        if not self.values:
            return MultiNumber([0])
        return MultiNumber(self.values)

    def compose(self, board) -> Dict:
        if len(self.values) <= 1:
            value = 0
            if len(self.values) == 1:
                value = self.values[0]
            return get_col(
                get_dummy(height=0.175),
                get_text(str(value)),
                get_dummy(height=0.175),
            )
        if len(self.values) == 2:
            text_a = get_text(str(self.values[0]))
            text_b = get_text(str(self.values[1]))
            return get_col(
                get_dummy(height=0.175),
                get_row(
                    text_a,
                    text_b
                ),
                get_dummy(height=0.175),
            )
        elif len(self.values) == 3:
            text_a = get_text(str(self.values[0]))
            text_b = get_text(str(self.values[1]))
            text_c = get_text(str(self.values[2]))
            return get_col(
                get_row(
                    text_a,
                    text_b,
                    # spacing=0
                ),
                text_c,
            )
        elif len(self.values) == 4:
            text_a = get_text(str(self.values[0]))
            text_b = get_text(str(self.values[1]))
            text_c = get_text(str(self.values[2]))
            text_d = get_text(str(self.values[3]))
            return get_col(
                get_row(
                    text_a,
                    text_b,
                ),
                get_row(
                    text_c,
                    text_d
                )
            )
        else:
            # 我也不知道为什么会出现>5个数字的情况
            return get_text("")

    @classmethod
    def type(cls) -> bytes:
        return Rule1W.name[0].encode("ascii")

    def code(self) -> bytes:
        return encode(self.values)

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        var_list = board.batch([
            self.pos.right(), self.pos.right().down(),
            self.pos.down(), self.pos.left().down(),
            self.pos.left(), self.pos.left().up(),
            self.pos.up(), self.pos.right().up()
        ], mode="variable")

        possible_list = [[]]

        for value in MineStatus_1W(self.values):
            bool_list = [(value >> i) & 1 == 1 for i in reversed(range(8))]
            flag = False
            for index, var in enumerate(var_list):
                if var is None and bool_list[index]:
                    flag = True
                    break
                if var is None:
                    continue
                possible_list[-1].append(bool_list[index])
            if flag:
                possible_list.pop(-1)
            possible_list.append([])

        if any(v is None for v in var_list):
            var_list = [var for var in var_list if var is not None]
        possible_list.pop(-1)

        if possible_list:
            model.AddAllowedAssignments(var_list, possible_list).OnlyEnforceIf(s)
        else:
            model.Add(sum(var_list) == 0).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/28 18:21
# @Author  : Wu_RH
# @FileName: 3N.py
"""
[4N]邻居：将每个雷向最近的雷连有向边（如果离一个雷最近的雷有多个，就都连），与一个雷连接的边竖无限制，雷线索表示从该雷开始走，只能沿着有向边走，走到已经走过的雷经过的边数的最小值。
"""



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/15 11:28
# @Author  : Wu_RH
# @FileName: RuleC.py
"""
[RULE-011]没想好: 线索表示周围八格中周围四格没有雷的雷数。
"""
from ortools.sat.python.cp_model import IntVar, CpModel

from minesweepervariants.abs.Rrule import AbstractClueRule, AbstractClueValue
from minesweepervariants.abs.board import AbstractBoard, AbstractPosition
from minesweepervariants.impl.summon.solver import Switch


class RuleC(AbstractClueRule):
    name = ["Rule-011", "没想好", "τB", "RULE-011"]
    doc = "线索表示周围八格中周围四格没有雷的雷数。"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        mines_map = {}
        for pos, _ in board("F"):
            mines_map[pos] = board.batch(pos.neighbors(1, 1), "type").count("F") == 0
        for pos, _ in board("N"):
            value = 0
            for _pos in pos.neighbors(2):
                if _pos not in mines_map:
                    continue
                value += mines_map[_pos]
            board[pos] = ValueC(pos, bytes([value]))
        return board

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()

        var_map = {}
        for pos, var in board(mode="var"):
            boolVar = model.NewBoolVar(f"bool[{pos}]")
            sum_var = sum(board.batch(pos.neighbors(1, 1), "var", drop_none=True))
            model.Add(sum_var == 0).OnlyEnforceIf([boolVar, var])
            model.Add(sum_var != 0).OnlyEnforceIf([boolVar.Not(), var])
            model.Add(boolVar == 0).OnlyEnforceIf(var.Not())
            var_map[pos] = boolVar
        for pos, obj in board():
            if not isinstance(obj, ValueC):
                continue
            var_list = [var_map[_pos] for _pos in pos.neighbors(2) if _pos in var_map]
            obj.create_constraints_(model, var_list, switch.get(model, pos))


class ValueC(AbstractClueValue):

    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos, code)
        self.value = code[0]

    def __repr__(self) -> str:
        return str(self.value)

    def code(self) -> bytes:
        return bytes([self.value])

    @classmethod
    def type(cls) -> bytes:
        return RuleC.name[0].encode()

    def create_constraints_(self, model: CpModel, var_list: list, switch: 'IntVar'):
        model.Add(sum(var_list) == self.value).OnlyEnforceIf(switch)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/17 08:52
# @Author  : xxx
# @FileName: generate_game.py
"""
普通模式-随机生成题板
"""
import base64
import os
import time

from minesweepervariants.impl.impl_obj import ModelGenerateError, get_board, encode_board
from minesweepervariants.impl.summon import Summon
from minesweepervariants.impl.summon.game import GameSession, PUZZLE
from minesweepervariants.impl.summon.summon import GenerateError
from minesweepervariants.utils.impl_obj import get_seed
from minesweepervariants.utils.tool import get_logger, get_random

from minesweepervariants.config.config import DEFAULT_CONFIG, PUZZLE_CONFIG

# ==== 获取默认值 ====
CONFIG = {}
CONFIG.update(DEFAULT_CONFIG)
CONFIG.update(PUZZLE_CONFIG)


def main(
        log_lv: str,  # 日志等级
        seed: int,  # 随机种子
        size: tuple[int, int],  # 题板尺寸
        total: int,  # 总雷数
        rules: list[str],  # 所有的规则集合
        query: int,  # 至少有几线索推理
        attempts: int,  # 尝试次数
        dye: str,  # 染色规则
        mask_dye: str,   # 异形题板
        drop_r: bool,  # 在推理时候是否隐藏R推理
        early_stop: bool,  # 如果达到指定的线索数量 直接跳出
        board_class: str,  # 题板的类名
        vice_board: bool,  # 启用删除副板
        unseed: bool  # 是否禁用种子来快速生成题目
):
    rule_code = rules[:]
    logger = get_logger(log_lv=log_lv)
    get_random(seed, new=True)
    s = Summon(size=size, total=total, rules=rules, board=board_class,
               drop_r=drop_r, mask=mask_dye, dye=dye, vice_board=vice_board)

    rule_text = ""
    for rule in rules:
        rule_text += "[" + (rule.split(CONFIG['delimiter'])[0] if
                            CONFIG['delimiter'] in rule else rule) + "]"
    if rule_text == "":
        rule_text = "[V]"
    if dye:
        rule_text += f"[{dye}]"
    if mask_dye:
        rule_text += f"[{mask_dye}]"

    size_a = 0
    size_b = 0
    size_c = len(s.board.get_interactive_keys())
    for key in s.board.get_interactive_keys():
        bound = s.board.boundary(key)
        size_a = max(size_a, bound.x + 1)
        size_b = max(size_b, bound.y + 1)
    rule_text += f"{size[0]}x{size[1]}"
    if size_c > 1:
        rule_text += f"x{size_c}"
    total = s.total
    logger.info(f"total mines: {total}")

    attempt_index = 0

    while True:
        s = Summon(size=size, total=total, rules=rule_code[:], board=board_class,
                   drop_r=drop_r, mask=mask_dye, dye=dye, vice_board=vice_board)
        if unseed:
            s.unseed = True
        get_random(seed, new=True)
        a_time = time.time()
        if attempts != -1 and attempt_index >= attempts:
            break
        attempt_index += 1
        game = GameSession(s, mode=PUZZLE, drop_r=drop_r)
        try:
            game.board = s.create_puzzle()
        except ModelGenerateError:
            continue
        except GenerateError:
            continue
        game.answer_board = get_board(board_class)(code=s.answer_board_code)
        # game.board = get_board("0B")(code=b'')
        # game.answer_board = get_board("0B")(code=b'')
        _board = game.board.clone()
        game.logger.info("\n" + "=" * 30 + "\nanswer_board:\n" + game.answer_board.show_board())
        game.logger.info("\nboard:\n" + game.board.show_board())
        game.logger.info("board: " + str(game.board.encode()))
        game.logger.info("answer: " + str(game.answer_board.encode()))
        try:
            clue_freq = game.check_difficulty(query, br=early_stop)
        except ModelGenerateError:
            continue
        if not clue_freq:
            logger.warn("生成失败")
            continue
        game.logger.info("\n" + "=" * 50 + "\n" + str(clue_freq))
        time_used = time.time() - a_time
        n_num = len([None for _ in _board("N")])
        board_str = _board.show_board()
        answer = game.answer_board.show_board()
        board_code = _board.encode()
        answer_code = game.answer_board.encode()

        if not any(k >= query for k in clue_freq):
            continue

        if not os.path.exists(CONFIG["output_path"]):
            os.makedirs(CONFIG["output_path"])

        mask = 0
        for key in _board.get_board_keys():
            for _, obj in _board(key=key):
                if obj is None:
                    mask += 1
                mask <<= 1

        # 计算需要的字节长度
        byte_length = (mask.bit_length() + 7) // 8  # 计算所需字节数
        byte_length = max(byte_length, 1)  # 确保至少 1 字节

        mask = mask.to_bytes(byte_length, "big", signed=False)

        with (open(os.path.join(CONFIG["output_path"], "demo.txt"), "a", encoding="utf-8") as f):
            f.write("\n" + ("=" * 100) + "\n\n生成时间" + logger.get_time() + "\n")
            f.write(f'线索表\n')
            if 0 in clue_freq:
                f.write("存在线索数为0的可推格 请注意辨别\n")
            if early_stop:
                f.write("(已启用-e参数 线索图不准确)\n")
            f.write(f'{clue_freq}\n')
            f.write(f"生成用时:{time_used}s\n")
            f.write(f"总雷数: {total}/{n_num}\n")
            f.write(f"种子: {get_seed()}\n")
            f.write("\n" + rule_text)
            f.write("\n" + board_str)
            f.write("\n" + answer)

            f.write(f"\n答案: img -c {encode_board(answer_code)} ")
            f.write(f"-r \"{rule_text}-R{total}/")
            f.write(f"{n_num}")
            if unseed:
                f.write(f"-{get_seed()}\" ")
            else:
                f.write(" ")
            f.write("-o answer\n")

            f.write(f"\n题板: img -c {encode_board(board_code)} ")
            f.write(f"-r \"{rule_text}-R{'*' if drop_r else total}/")
            f.write(f"{n_num}")
            if unseed:
                f.write(f"-{get_seed()}\" ")
            else:
                f.write(" ")
            f.write("-o demo\n")

            rule_code = [base64.urlsafe_b64encode(rule.encode("utf-8")).decode("utf-8") for rule in rule_code]
            f.write(f"\n题板代码: \n{encode_board(answer_code)}:{mask.hex()}:{':'.join(rule_code)}\n")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/10 09:52
# @Author  : xxx
# @FileName: c.py

from . import AbstractDye


class DyeC(AbstractDye):
    name = "nq" # no quadrant
    __doc__ = "无方染色"

    def dye(self, board):
        dye = True
        for key in board.get_interactive_keys():
            dye = not dye
            for pos, _ in board(key=key):
                _dye = dye ^ ((pos.x % 2 + pos.y % 2) > 1)
                board.set_dyed(pos, _dye)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/28 15:59
# @Author  : Wu_RH
# @FileName: Gd.py
"""
[Gd]: 每列雷数不少于左一列雷数；每行雷数不少于上一行雷数
"""

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class RuleGd(AbstractMinesRule):
    name = ["Gd", "GD"]
    doc = "1.每行值不大于上一行 2.每列值不大于左一列"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s1 = switch.get(model, self)
        s2 = switch.get(model, self)
        for key in board.get_interactive_keys():
            boundary_pos = board.boundary(key=key)

            col_positions = board.get_col_pos(boundary_pos)
            col_sums = [
                sum(board.get_variable(_pos) for _pos in board.get_row_pos(pos))
                for pos in col_positions
            ]
            # 所有 col_sums 相等
            for i in range(1, len(col_sums)):
                model.Add(col_sums[i - 1] <= col_sums[i]).OnlyEnforceIf(s1)

            row_positions = board.get_row_pos(boundary_pos)
            row_sums = [
                sum(board.get_variable(_pos) for _pos in board.get_col_pos(pos))
                for pos in row_positions
            ]
            # 所有 row_sums 相等
            for i in range(1, len(row_sums)):
                model.Add(row_sums[i - 1] <= row_sums[i]).OnlyEnforceIf(s2)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/04 07:39
# @Author  : Wu_RH
# @FileName: 4V.py
"""
[4V']2X'plusplus: 线索表示数字是三个题板中相同位置的其中某两个范围中心3*3区域的雷总数
"""
from typing import Dict

from minesweepervariants.utils.web_template import MultiNumber
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition, MASTER_BOARD
from ....utils.image_create import get_text, get_row
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG
from ....utils.tool import get_random

NAME_4Vp = ["4Va", "4Vb"]


class Rule4Vp(AbstractClueRule):
    name = ["4V'", "映射'"]
    doc = "线索表示数字是三个题板中相同位置的其中某两个范围中心3*3区域的雷总数"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        size = (board.boundary().x + 1, board.boundary().y + 1)
        board.generate_board(NAME_4Vp[0], size)
        board.generate_board(NAME_4Vp[1], size)
        board.set_config(NAME_4Vp[0], "interactive", True)
        board.set_config(NAME_4Vp[1], "interactive", True)
        board.set_config(NAME_4Vp[0], "row_col", True)
        board.set_config(NAME_4Vp[1], "row_col", True)
        board.set_config(NAME_4Vp[0], "VALUE", VALUE_QUESS)
        board.set_config(NAME_4Vp[1], "VALUE", VALUE_QUESS)
        board.set_config(NAME_4Vp[0], "MINES", MINES_TAG)
        board.set_config(NAME_4Vp[1], "MINES", MINES_TAG)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()

        for key in [MASTER_BOARD] + NAME_4Vp:
            for pos, _ in board("N", key=key):
                neighbors_list = []
                for _key in [MASTER_BOARD] + NAME_4Vp:
                    _pos = pos.clone()
                    _pos.board_key = _key
                    neighbors_list.append(_pos.neighbors(0, 2))
                value_a, value_b = random.sample([board.batch(positions, mode="type").count("F")
                                                  for positions in neighbors_list], 2)
                value = Value4Vp(pos=pos, code=bytes([value_b << 4 | value_a]))
                board.set_value(pos, value)

        return board


class Value4Vp(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.neighbors_list = []
        for key in [MASTER_BOARD] + NAME_4Vp:
            _pos = pos.clone()
            _pos.board_key = key
            self.neighbors_list.append(_pos.neighbors(0, 2))
        self.value_a = code[0] & 0xf
        self.value_b = code[0] >> 4
        self.pos = pos

    @classmethod
    def method_choose(cls) -> int:
        return 1

    @classmethod
    def type(cls) -> bytes:
        return Rule4Vp.name[0].encode("ascii")

    def __repr__(self) -> str:
        value = [self.value_a, self.value_b]
        value.sort()
        return f"{value[0]} {value[1]}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return (self.neighbors_list[0] +
                self.neighbors_list[1] +
                self.neighbors_list[2])

    def web_component(self, board) -> Dict:
        value = [self.value_a, self.value_b]
        value.sort()
        return MultiNumber(value)

    def compose(self, board) -> Dict:
        value = [self.value_a, self.value_b]
        value.sort()
        text_a = get_text(str(value[0]))
        text_b = get_text(str(value[1]))
        return get_row(
            text_a,
            text_b
        )

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        # 收集三组邻居的有效位置
        groups = []
        for i, neighbor in enumerate(self.neighbors_list):
            var_list = board.batch(neighbor, mode="variable", drop_none=True)
            if var_list:
                groups.append({
                    "index": i,
                    "vars": var_list,
                    "sum": sum(var_list)
                })

        num_groups = len(groups)

        # 如果有效组少于2个，无法满足要求（添加不可能约束）
        if num_groups < 2:
            model.AddBoolOr([])
            return

        # 创建选择变量：选择哪两组分别对应值A和值B
        a_group = model.NewIntVar(0, num_groups - 1, f"[{Rule4Vp.name}]a_group")
        b_group = model.NewIntVar(0, num_groups - 1, f"[{Rule4Vp.name}]b_group")

        # 确保选择的组不同
        model.Add(a_group != b_group).OnlyEnforceIf(s)

        # 为每个组创建指示变量
        group_flags = []
        for i, group in enumerate(groups):
            # 创建指示变量：当前组是否被选为值A的组
            is_a_group = model.NewBoolVar(f"is_a_group_{i}")
            model.Add(a_group == i).OnlyEnforceIf([is_a_group, s])
            model.Add(a_group != i).OnlyEnforceIf([is_a_group.Not(), s])

            # 创建指示变量：当前组是否被选为值B的组
            is_b_group = model.NewBoolVar(f"is_b_group_{i}")
            model.Add(b_group == i).OnlyEnforceIf([is_b_group, s])
            model.Add(b_group != i).OnlyEnforceIf([is_b_group.Not(), s])

            # 添加约束：如果被选为值A的组，则必须等于值A
            model.Add(group["sum"] == self.value_a).OnlyEnforceIf([is_a_group, s])

            # 添加约束：如果被选为值B的组，则必须等于值B
            model.Add(group["sum"] == self.value_b).OnlyEnforceIf([is_b_group, s])

            group_flags.append((is_a_group, is_b_group))

        # 确保恰好一组被选为值A，一组被选为值B
        model.Add(sum([a for a, _ in group_flags]) == 1).OnlyEnforceIf(s)
        model.Add(sum([b for _, b in group_flags]) == 1).OnlyEnforceIf(s)

        # 第三组没有约束（可以任意值）

    def code(self) -> bytes:
        return bytes([self.value_b << 4 | self.value_a])



"""
[3D] 辞典：所有雷从左到右，从上到下依次标号。线索表示周围八格的雷的标号之和
"""

from ....abs.board import AbstractBoard, AbstractPosition
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....utils.tool import get_logger
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG


def encode_int_7bit(n: int) -> bytes:
    # 编码主体：每7位 -> 1字节（bit6~bit0，bit7=0）
    if n == 0:
        return b'\x00'
    payload = []

    while n > 0:
        payload.append(n & 0x7f)
        n >>= 7

    return bytes(payload)


def decode_bytes_7bit(data: bytes) -> int:
    if len(data) == 0:
        return 0

    result = 0
    for i in data[::-1]:
        result <<= 7
        result |= i

    return result


class Rule3D(AbstractClueRule):
    name = ["3D", "辞典", "Dict"]
    doc = "所有雷从左到右，从上到下依次标号。线索表示周围八格的雷的标号之和"

    def fill(self, board: AbstractBoard) -> AbstractBoard:
        logger = get_logger()
        for key in board.get_interactive_keys():
            dict = {}
            x = board.boundary(key).x + 1
            y = board.boundary(key).y + 1
            for i in range(x):
                for j in range(y):
                    pos = board.get_pos(i, j, key)
                    if board.get_type(pos) == "F":
                        dict[pos] = len(dict) + 1
            if not dict:
                logger.error(f"[3D] R = 0")
                return board
            for pos, _ in board("N", key=key):
                value = 0
                for nei in pos.neighbors(2):
                    if board.get_type(nei) == "F":
                        value += dict[nei]
                board.set_value(pos, Value3D(pos, code=encode_int_7bit(value)))
                logger.debug(f"[3D] put {value} to {pos}")
        return board


class Value3D(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos)
        self.value = decode_bytes_7bit(code)
        self.neighbors = pos.neighbors(2)

    def __repr__(self) -> str:
        return f"{self.value}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbors

    @classmethod
    def type(cls) -> bytes:
        return Rule3D.name[0].encode("ascii")

    def code(self) -> bytes:
        return encode_int_7bit(self.value)

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        x = board.boundary(self.pos.board_key).x + 1
        y = board.boundary(self.pos.board_key).y + 1
        N = x * y
        m = [board.get_variable(board.get_pos(i, j, self.pos.board_key)) for i in range(x) for j in range(y)]      # 雷格变量
        P = [model.NewIntVar(0, N, f"[3D]P{i}") for i in range(N)]                                                 # 前缀雷计数
        L = [model.NewIntVar(0, N, f"[3D]L{i}") for i in range(N)]                                                 # 该格的雷编号，非雷为 0

        # 连接 P 与 m（P 是 m 的前缀和）
        # P[0] = m[0]; P[i] = P[i-1] + m[i]
        model.Add(P[0] == m[0]).OnlyEnforceIf(s)
        for i in range(1, N):
            model.Add(P[i] == P[i-1] + m[i]).OnlyEnforceIf(s)

        # 若 m[i]=1，则 L[i]=P[i]；若 m[i]=0，则 L[i]=0
        for i in range(N):
            model.Add(L[i] == P[i]).OnlyEnforceIf([m[i], s])
            model.Add(L[i] == 0).OnlyEnforceIf([m[i].Not(), s])

        model.Add(sum(L[nei.x * y + nei.y] for nei in self.neighbors if board.in_bounds(nei)) == self.value).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/28 11:12
# @Author  : Wu_RH
# @FileName: BM.py
"""
[BM]多雷平衡(扫联会2)：每行每列的“多雷雷值”相同。(线索格不一定是多雷规则)
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class RuleBM(AbstractMinesRule):
    name = ["BM", "1B1M", "多雷平衡", "多雷平衡(扫联会2)"]
    doc = "染色格被视为两个雷, 每行每列的“多雷雷值”相同。(线索格不一定是多雷规则) (1).列平衡, (2).行平衡"

    def __init__(self, board: AbstractBoard, data=None):
        super().__init__(board, data)
        self.R_1M = None if data is None else int(data)

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s1 = switch.get(model, self)
        s2 = switch.get(model, self)
        for key in board.get_interactive_keys():
            boundary_pos = board.boundary(key=key)

            row_positions = board.get_row_pos(boundary_pos)
            row_sums = [
                sum(board.get_variable(_pos) * (2 if board.get_dyed(_pos) else 1)
                    for _pos in board.get_col_pos(pos))
                for pos in row_positions
            ]
            # 所有 row_sums 相等
            for i in range(1, len(row_sums)):
                model.Add(row_sums[i] == row_sums[0]).OnlyEnforceIf(s1)
            if self.R_1M:
                model.Add(row_sums[0] == (self.R_1M // (boundary_pos.y+1))).OnlyEnforceIf(s1)

            col_positions = board.get_col_pos(boundary_pos)
            col_sums = [
                sum(board.get_variable(_pos) * (2 if board.get_dyed(_pos) else 1)
                    for _pos in board.get_row_pos(pos))
                for pos in col_positions
            ]
            # 所有 col_sums 相等
            for i in range(1, len(col_sums)):
                model.Add(col_sums[i] == col_sums[0]).OnlyEnforceIf(s2)
            if self.R_1M:
                model.Add(col_sums[0] == (self.R_1M // (boundary_pos.x+1))).OnlyEnforceIf(s2)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/10 11:44
# @Author  : xxx
# @FileName: Q1.py

"""
[3D1Q]无方: 每个2x2x2区域内都至少有2个雷
"""

from typing import List

from .. import Abstract3DMinesRule
from .....abs.board import AbstractPosition, AbstractBoard


def block(a_pos: AbstractPosition, board: AbstractBoard) -> List[AbstractPosition]:
    b_pos = a_pos.up()
    c_pos = a_pos.left()
    d_pos = b_pos.left()
    if not board.in_bounds(d_pos):
        return []
    e_pos = a_pos.clone()
    e_pos = Rule1Q.up(board, e_pos)
    if e_pos is None:
        return []
    f_pos = e_pos.up()
    g_pos = e_pos.left()
    h_pos = f_pos.left()
    if not board.in_bounds(h_pos):
        return []
    return [a_pos, b_pos, c_pos, d_pos, e_pos, f_pos, g_pos, h_pos]


class Rule1Q(Abstract3DMinesRule):
    name = ["3D1Q", "3DQ", "三维无方"]
    doc = "每个2x2x2区域内都至少有2个雷"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        for pos, _ in board():
            if not (pos_block := block(pos, board)):
                continue
            var_list = [board.get_variable(pos) for pos in pos_block]
            model.Add(sum(var_list) > 1).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
[1E] 视野 (Eyesight)：线索表示四方向上能看到的非雷格数量（包括自身），雷会阻挡视线
"""
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition


class Rule1E(AbstractClueRule):
    name = ["1E", "E", "视野", "Eyesight"]
    doc = "线索表示四方向上能看到的非雷格数量（包括自身），雷会阻挡视线"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            value = 1  # 包括自身
            # 四方向的函数
            direction_funcs = [
                lambda _n: type(pos)(pos.x + _n, pos.y, pos.board_key),  # 右
                lambda _n: type(pos)(pos.x - _n, pos.y, pos.board_key),  # 左
                lambda _n: type(pos)(pos.x, pos.y + _n, pos.board_key),  # 上
                lambda _n: type(pos)(pos.x, pos.y - _n, pos.board_key)   # 下
            ]

            for fn in direction_funcs:
                n = 1
                while True:
                    next_pos = fn(n)
                    if not board.in_bounds(next_pos):
                        break
                    if board.get_type(next_pos) == "F":  # 遇到雷，视线被阻挡
                        break
                    value += 1
                    n += 1

            obj = Value1E(pos, bytes([value]))
            board.set_value(pos, obj)
        return board


class Value1E(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.pos = pos

    def __repr__(self):
        return str(self.value)

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        positions = []
        for i in [
            (1, 0), (0, 1),
            (-1, 0), (0, -1),
        ]:
            n = 0
            while board.get_type(pos := self.pos.shift(i[0] * n, i[1] * n)) not in "F":
                n += 1
                positions.append(pos)
        return positions

    @classmethod
    def type(cls) -> bytes:
        return Rule1E.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        def dfs(value: int, index=0, info: dict = None):
            if info is None:
                info = {"T": [], "F": []}
            if index == 4:
                if value == 1:
                    possible_list.append((set(info["T"]), [var for var in info["F"] if var is not None]))
                return

            # 四方向的函数
            direction_funcs = [
                lambda n: type(self.pos)(self.pos.x + n, self.pos.y, self.pos.board_key),  # 右
                lambda n: type(self.pos)(self.pos.x - n, self.pos.y, self.pos.board_key),  # 左
                lambda n: type(self.pos)(self.pos.x, self.pos.y + n, self.pos.board_key),  # 上
                lambda n: type(self.pos)(self.pos.x, self.pos.y - n, self.pos.board_key)   # 下
            ]

            fn = direction_funcs[index]

            for i in range(value):
                current_pos = fn(i)
                if not board.in_bounds(current_pos):
                    dfs(value - i, index + 1, info)
                    break

                _var_t = board.get_variable(current_pos)
                if _var_t is None:
                    dfs(value - i, index + 1, info)
                    break

                next_pos = fn(i + 1)
                _var_f = board.get_variable(next_pos) if board.in_bounds(next_pos) else None

                info["T"].append(_var_t)
                info["F"].append(_var_f)
                dfs(value - i, index + 1, info)
                info["F"].pop(-1)

            for i in range(value):
                current_pos = fn(i)
                if not board.in_bounds(current_pos):
                    continue
                if board.get_variable(current_pos) is None:
                    continue
                info["T"].pop(-1)

        model = board.get_model()
        s = switch.get(model, self)
        possible_list = []

        dfs(value=self.value)
        tmp_list = []

        for vars_t, vars_f in possible_list:
            tmp = model.NewBoolVar("tmp")
            model.Add(sum(vars_t) == 0).OnlyEnforceIf(tmp)
            if vars_f and any(var is not None for var in vars_f):
                model.AddBoolAnd([var for var in vars_f if var is not None]).OnlyEnforceIf(tmp)
            tmp_list.append(tmp)

        if tmp_list:
            model.AddBoolOr(tmp_list).OnlyEnforceIf(s)
